"""Interfaces to Laplace operators."""

from bempp_cl.api.operators.boundary import common as _common


def single_layer(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Laplace single-layer boundary operator."""
    return _common.create_operator(
        "laplace_single_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [],
        "laplace_single_layer",
        "default_scalar",
        device_interface,
        precision,
        False,
    )


def double_layer(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Laplace double-layer boundary operator."""
    return _common.create_operator(
        "laplace_double_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [],
        "laplace_double_layer",
        "default_scalar",
        device_interface,
        precision,
        False,
    )


def adjoint_double_layer(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Laplace adjoint double-layer boundary operator."""
    return _common.create_operator(
        "laplace_adjoint_double_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [],
        "laplace_adjoint_double_layer",
        "default_scalar",
        device_interface,
        precision,
        False,
    )


def hypersingular(
    domain,
    range_,
    dual_to_range,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Laplace hypersingular boundary operator."""
    if domain.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain shapeset must be of type 'p1_discontinuous'.")

    if dual_to_range.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Dual to range shapeset must be of type 'p1_discontinuous'.")

    return _common.create_operator(
        "laplace_hypersingular_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [],
        "laplace_single_layer",
        "laplace_hypersingular",
        device_interface,
        precision,
        False,
    )


def multitrace_operator(
    grid,
    target=None,
    space_type="p1",
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """
    Simplified version of multitrace operator assembly.

    Parameters
    ----------
    grid : Grid
        Bempp grid object.
    target : Grid
        The grid for the range spaces. If target is None then
        target is set to the input grid (that is the domain
        grid).
    space_type : string
        Controls which discretisation spaces are used.
        Supported values:
        - "p1", which means that all operators are discretised
        with P1 basis functions.
        - "p1-dp0", which means that P1 basis functions are
        used for the first row and column, and DP0 for the second.
        - "p1-dual0", which means that P1 basis functions are
        used for the first row and column, and DUAL0 for the second.
    parameters : Parameters
        An optional parameters object.
    assembler : string
        The assembler type.
    device_interface : DeviceInterface
        The device interface object to be used.
    precision : string
        Either "single" or "double" for single or
        double precision mode.

    Output
    ------
    The Helmholtz multitrace operator of the form
    [[-dlp, slp], [hyp, adj_dlp]], where
    dlp : double layer boundary operator
    slp : single layer boundary operator
    hyp : hypersingular boundary operator
    adj_dlp : adjoint double layer boundary operator.

    """
    import bempp_cl.api
    from bempp_cl.api.assembly.blocked_operator import BlockedOperator

    if space_type == "p1":
        space1 = bempp_cl.api.function_space(grid, "P", 1)
        space0 = space1
        if target is not None:
            target_space1 = bempp_cl.api.function_space(target, "P", 1)
            target_space0 = target_space1
            target_space_dual1 = target_space1
            target_space_dual0 = target_space1
        else:
            target_space1 = space1
            target_space0 = space1
            target_space_dual1 = space1
            target_space_dual0 = space1
    elif space_type == "p1-dp0":
        space1 = bempp_cl.api.function_space(grid, "P", 1)
        space0 = bempp_cl.api.function_space(grid, "DP", 0)
        if target is not None:
            target_space1 = bempp_cl.api.function_space(target, "P", 1)
            target_space0 = bempp_cl.api.function_space(target, "DP", 0)
            target_space_dual1 = target_space1
            target_space_dual0 = target_space0
    elif space_type == "p1-dual":
        space1 = bempp_cl.api.function_space(grid, "P", 1)
        space0 = bempp_cl.api.function_space(grid, "DUAL", 0)
        if target is not None:
            target_space1 = bempp_cl.api.function_space(target, "P", 1)
            target_space0 = bempp_cl.api.function_space(target, "DUAL", 0)
            target_space_dual1 = target_space1
            target_space_dual0 = target_space0
    else:
        raise ValueError(f"Unknown space type: {space_type}")

    slp = single_layer(
        space0,
        target_space1,
        target_space_dual0,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    dlp = double_layer(
        space1,
        target_space1,
        target_space_dual0,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    hyp = hypersingular(
        space1,
        target_space0,
        target_space_dual1,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    adj_dlp = adjoint_double_layer(
        space0,
        target_space0,
        target_space_dual1,
        parameters=parameters,
        assembler=assembler,
        device_interface=device_interface,
        precision=precision,
    )

    blocked = BlockedOperator(2, 2)

    blocked[0, 0] = -dlp
    blocked[0, 1] = slp
    blocked[1, 0] = hyp
    blocked[1, 1] = adj_dlp

    return blocked



"""Global initialization for Bempp."""

# # Monkey patch Numba to emit log messages when compiling

# import numba

# oldcompile = numba.core.registry.CPUDispatcher.compile


# def compile_with_log(*args, **kwargs):
# """Numba compilation with log messages."""
# import bempp_cl.api

# fun_name = args[0].py_func.__name__
# bempp_cl.api.log(f"Compiling {fun_name} for signature {args[1]}.", level="debug")
# res = oldcompile(*args, **kwargs)
# bempp_cl.api.log(f"Compilation finished.", level="debug")
# return res


# numba.core.registry.CPUDispatcher.compile = compile_with_log

import os as _os
import tempfile as _tempfile
import logging as _logging
import time as _time
import platform as _platform

from bempp_cl.api.utils import DefaultParameters
from bempp_cl.api.utils.helpers import MemProfiler

from bempp_cl.api.utils.helpers import assign_parameters
from bempp_cl.api.grid.io import import_grid
from bempp_cl.api.grid.io import export
from bempp_cl.api.grid.grid import Grid
from bempp_cl.api.assembly.grid_function import GridFunction
from bempp_cl.api.assembly.grid_function import real_callable
from bempp_cl.api.assembly.grid_function import complex_callable
from bempp_cl.api.assembly.grid_function import callable

from bempp_cl.api.space import function_space

from bempp_cl.api import shapes
from bempp_cl.api import integration
from bempp_cl.api import operators
from bempp_cl.api.linalg.direct_solvers import lu, compute_lu_factors
from bempp_cl.api.linalg.iterative_solvers import gmres, cg
from bempp_cl.api.assembly.discrete_boundary_operator import as_matrix
from bempp_cl.api.assembly.boundary_operator import ZeroBoundaryOperator
from bempp_cl.api.assembly.boundary_operator import MultiplicationOperator
from bempp_cl.api.assembly.blocked_operator import BlockedOperator
from bempp_cl.api.assembly.blocked_operator import GeneralizedBlockedOperator

from bempp_cl.api.fmm.fmm_assembler import clear_fmm_cache

from bempp_cl.api.utils import pool
from bempp_cl.api.utils.pool import create_device_pool

from numba.core.errors import (
    NumbaDeprecationWarning,
    NumbaPendingDeprecationWarning,
    NumbaPerformanceWarning,
)

import warnings

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaPendingDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaPerformanceWarning)
warnings.filterwarnings("ignore", message="splu requires CSC matrix format")

# Try importing OpenCL routines

try:
    from bempp_cl.core.opencl_kernels import set_default_cpu_device
    from bempp_cl.core.opencl_kernels import set_default_cpu_device_by_name
    from bempp_cl.core.opencl_kernels import set_default_gpu_device_by_name
    from bempp_cl.core.opencl_kernels import set_default_gpu_device
except ImportError:
    pass


CONSOLE_LOGGING_HANDLER = None
DEFAULT_LOGGING_FORMAT = "%(name)s:%(levelname)s: %(message)s"
DEBUG = _logging.DEBUG
TIMING = 11
INFO = _logging.INFO
WARNING = _logging.WARNING
ERROR = _logging.ERROR
CRITICAL = _logging.CRITICAL

LOG_LEVEL = {
    "debug": DEBUG,
    "timing": TIMING,
    "info": INFO,
    "warning": WARNING,
    "error": ERROR,
    "critical": CRITICAL,
}

GLOBAL_PARAMETERS = DefaultParameters()


def _init_logger():
    """Initialize the Bempp logger."""
    _logging.addLevelName(11, "TIMING")
    logger = _logging.getLogger("bempp")
    logger.setLevel(DEBUG)
    logger.addHandler(_logging.NullHandler())
    return logger


def log(message, level="info", flush=True):
    """Log including default flushing for IPython."""
    LOGGER.log(LOG_LEVEL[level], message)
    if flush:
        flush_log()


def flush_log():
    """Flush all handlers. Necessary for Jupyter."""
    for handler in LOGGER.handlers:
        handler.flush()


def enable_console_logging(level="info"):
    """Enable console logging and return the console handler."""
    # pylint: disable=W0603
    global CONSOLE_LOGGING_HANDLER
    if not CONSOLE_LOGGING_HANDLER:
        console_handler = _logging.StreamHandler()
        console_handler.setLevel(LOG_LEVEL[level])
        if pool.is_worker():
            console_handler.setFormatter(
                _logging.Formatter(f"%(name)s:PROC{pool._MY_ID}:%(levelname)s: %(message)s", "%H:%M:%S")
            )
        else:
            console_handler.setFormatter(_logging.Formatter("%(name)s:HOST:%(levelname)s: %(message)s", "%H:%M:%S"))
        LOGGER.addHandler(console_handler)
        CONSOLE_LOGGING_HANDLER = console_handler
    return CONSOLE_LOGGING_HANDLER


# def enable_file_logging(file_name, level=DEBUG, logging_format=DEFAULT_LOGGING_FORMAT):
# """Enable logging to a specific file."""

# file_handler = _logging.FileHandler(file_name)
# file_handler.setLevel(level)
# file_handler.setFormatter(_logging.Formatter(logging_format, "%H:%M:%S"))
# LOGGER.addHandler(file_handler)
# return file_handler


def set_logging_level(level):
    """Set the logging level."""
    LOGGER.setLevel(LOG_LEVEL[level])


# pylint: disable=too-few-public-methods
class Timer:
    """Context manager to measure time in Bempp."""

    def __init__(self, enable_log=True, message="", level="timing"):
        """Construct."""
        self.start = 0
        self.end = 0
        self.interval = 0
        self.enable_log = enable_log
        self.level = level
        self.message = message

    def __enter__(self):
        """Enter."""
        if self.enable_log:
            log("Start operation: " + self.message, level=self.level)
        self.start = _time.time()
        return self

    def __exit__(self, *args):
        """Exit."""
        self.end = _time.time()
        self.interval = self.end - self.start
        if self.enable_log:
            log(
                "Finished Operation: " + self.message + f": {self.interval}s",
                level=self.level,
            )


LOGGER = _init_logger()

BEMPP_PATH = _os.path.abspath(_os.path.join(_os.path.dirname(_os.path.realpath(__file__)), ".."))

# pylint: disable=W0702
# try:
#    if _os.environ['BEMPP_CONSOLE_LOGGING'] == '1':
#        enable_console_logging()
# except:
#    pass

TMP_PATH = _tempfile.mkdtemp()

# Get the path to Gmsh


def _gmsh_path():
    """Find Gmsh."""
    from bempp_cl.api.utils import which

    if _os.name == "nt":
        gmp = which("gmsh.exe")
        if gmp is None:
            gmp = which("gmsh")
    else:
        gmp = which("gmsh")
    if gmp is None:
        print("Could not find Gmsh." + "Interactive plotting and shapes module not available.")
    return gmp


def check_for_fmm():
    """Return true of compatible FMM found."""
    exafmm_found = False
    try:
        import exafmm
    except ImportError:
        exafmm_found = False
    else:
        exafmm_found = True

    return exafmm_found


def _get_version():
    """Get version string."""
    from bempp_cl import version

    return version.__version__


GMSH_PATH = _gmsh_path()

__version__ = _get_version()

PLOT_BACKEND = "gmsh"
try:
    ipy = get_ipython()
    if ipy.__class__.__name__ == "ZMQInteractiveShell":
        # We are in a jupyter notebook, so change plotting backend
        PLOT_BACKEND = "jupyter_notebook"
except NameError:
    pass

USE_JIT = True

CPU_OPENCL_DRIVER_FOUND = False
GPU_OPENCL_DRIVER_FOUND = False

if _platform.system() == "Darwin":
    DEFAULT_DEVICE_INTERFACE = "numba"
else:
    try:
        from bempp_cl.core.opencl_kernels import find_cpu_driver

        CPU_OPENCL_DRIVER_FOUND = find_cpu_driver()
    except:  # noqa: E722
        pass

    try:
        from bempp_cl.core.opencl_kernels import find_gpu_driver

        GPU_OPENCL_DRIVER_FOUND = find_gpu_driver()
    except:  # noqa: E722
        pass

    if CPU_OPENCL_DRIVER_FOUND:
        DEFAULT_DEVICE_INTERFACE = "opencl"
    else:
        DEFAULT_DEVICE_INTERFACE = "numba"

if DEFAULT_DEVICE_INTERFACE == "numba":
    log("Numba backend activated. For full performance the OpenCL backend with an OpenCL CPU driver is required.")

DEFAULT_PRECISION = "double"
VECTORIZATION_MODE = "auto"

BOUNDARY_OPERATOR_DEVICE_TYPE = "cpu"
POTENTIAL_OPERATOR_DEVICE_TYPE = "cpu"

ALL = -1  # Useful global identifier



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 30 11:32:22 2025

@author: tom
"""

import bempp_cl.api
from bempp_cl.api.operators.boundary import helmholtz, sparse
from bempp_cl.api.operators.potential import helmholtz as helmholtz_potential
from bempp_cl.api.assembly.discrete_boundary_operator import DiagonalOperator
from scipy.sparse.linalg import gmres as scipy_gmres
from bempp_cl.api.linalg import gmres
import numpy as np
from tqdm import tqdm
import warnings
import electroacPy.general as gtb
from .ACSHelpers_ import (getSurfaceAdmittance, mirror_mesh, 
                         greenMonopole, incidenceCoeff, buildGridFunction,
                         element2vertice_pressure)

warnings.filterwarnings("ignore", message="splu requires CSC matrix format")
warnings.filterwarnings("ignore", message="splu converted its input to CSC format")

try:
    from pyopencl import CompilerWarning
    warnings.filterwarnings("ignore", category=CompilerWarning)
except:
    None


class pointSource:
    def __init__(self, sourcePosition, volumeVelocity, frequency, c=343, 
                 rho=1.22, boundary_conditions=None, **kwargs):
        """
        Simulate point sources in free space.

        Parameters
        ----------
        sourcePosition : list of list
            Source(s) coordinates [[x1, y1, z1], [x2, y2, z2], ..., [xN, yN, zN]].
        volumeVelocity : list of numpy array
            Sources volume velocity.
        frequency : numpy array
            Range of study.
        c : float, optional
            Speed of sound in propagation medium. The default is 343 m/s (air).
        rho : float, optional
            Density of propagation medium. The default is 1.22 kg/m^3.
        boundary_conditions : boundaryConditions object, optional
            Set of conditions on boundaries / infinite boundaries. The default is None.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        
        self.xSource = np.array(sourcePosition)
        self.xSystem = None
        self.QSource = np.array(volumeVelocity)
        self.QSystem = None
        self.frequency = frequency
        self.k = self.frequency*2*np.pi / c
        self.c = c
        self.rho = rho
        self.boundary_conditions = boundary_conditions
        self.sizeFactor = 1
        
        self.Nsource = len(self.xSource)
        if self.boundary_conditions is not None:
            self.mirror_system()
        else:
            self.xSystem = self.xSource
            
        self.isComputed = False

        # compatibility layer for integration with already-existing BEM eval 
        # and plots        
        if "radiatingElement" in kwargs:
            self.radiatingElement = kwargs["radiatingElement"]
        self.meshPath = None


    def mirror_system(self):
        self.xSystem = np.copy(self.xSource)
        for param in self.boundary_conditions.parameters:
            if param in ["x", "X"]:
                xmirror = np.copy(self.xSystem)
                xmirror[:, 0] = 2*self.boundary_conditions.parameters[param]["offset"] - xmirror[:, 0]
                self.xSystem = np.concatenate([self.xSystem, xmirror])
                self.sizeFactor *= 2
            if param in ["y", "Y"]:
                ymirror = np.copy(self.xSystem)
                ymirror[:, 1] = 2*self.boundary_conditions.parameters[param]["offset"] - ymirror[:, 1]
                self.xSystem = np.concatenate([self.xSystem, ymirror])
                self.sizeFactor *= 2
            if param in ["z", "Z"]:
                zmirror = np.copy(self.xSystem)
                zmirror[:, 2] = 2*self.boundary_conditions.parameters[param]["offset"] - zmirror[:, 2]
                self.xSystem = np.concatenate([self.xSystem, zmirror])
                self.sizeFactor *= 2
        self.QSystem = np.tile(self.QSource, (self.sizeFactor, 1))

    def getMicPressure(self, xMic, individualSpeakers=True):
        """
        Compute pressure at evaluation points (xMic).

        Parameters
        ----------
        xMic : numpy array
            Evaluation points in 3D space: shape (Nmic, 3).

        Returns
        -------
        pmic : numpy array
            Evaluated pressure (complex). Shape (Nfft, Nmic, Nsource)

        """
        pmic = np.zeros([len(self.frequency), len(xMic), len(self.xSystem)],
                        dtype=complex) 
        for m, mic in enumerate(xMic):
            for s, source in enumerate(self.xSystem):
                R = ((source[0]-mic[0])**2 + 
                    (source[1]-mic[1])**2 + 
                    (source[2]-mic[2])**2)**0.5
                pmic[:, m, s] = (1j*self.k*self.rho*self.c*self.QSystem[s, :] * 
                                 np.exp(-1j*self.k*R) / (4*np.pi*R))
        
        # need to sum all corresponding image/source group
        if self.sizeFactor != 1:
            Nsources = len(self.xSource)
            Nimages = self.sizeFactor * Nsources - Nsources 
            for s in range(Nsources):
                for i in range(Nimages//2):
                    pmic[:, :, s] += pmic[:, :, s + Nsources*i]
                            
        pmic = pmic[:, :, :Nsources]
        
        if individualSpeakers is True:
            out = (np.sum(pmic, 2), pmic)
        elif individualSpeakers is False:
            out = np.sum(pmic, 2)
        return out


class pointSourceBEM:
    def __init__(self, meshPath, xSource, volumeVelocity, frequency, 
             domain="interior", c_0=343, rho_0=1.22, **kwargs):
        """
        Simulate point source with additional boundary-element mesh for diffraction. 
        Mostly useful for room acoustics.

        Parameters
        ----------
        meshPath : str
            Path to simulation mesh.
        xSource : list of list
            Source(s) coordinates.
        velocity : list of numpy array
            Source volume velocity.
        frequency : numpy array
            Range of study.
        domain : TYPE, optional
            DESCRIPTION. The default is "exterior".
        c_0 : TYPE, optional
            DESCRIPTION. The default is 343.
        rho_0 : TYPE, optional
            DESCRIPTION. The default is 1.22.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        
        # inputs
        if isinstance(xSource, np.ndarray) is True:
            self.xSource = xSource
        else:
            self.xSource = np.array(xSource)
        self.xSystem = None
        self.QSource = np.array(volumeVelocity)
        self.QSystem = None
        self.frequency = frequency
        self.domain = domain
        self.c_0 = c_0
        self.rho_0 = rho_0
        self.kwargs = kwargs

        # check if mesh is v2
        self.meshPath = gtb.geometry.check_mesh(meshPath)
        if "radiatingElement" in kwargs:
            self.radiatingElement = kwargs["radiatingElement"]
        
        # other
        self.k = self.frequency * 2 * np.pi / self.c_0
        self.Ns = len(self.xSource)
        self.isComputed = False
        
        # initialize possible boundary conditions
        self.impedanceSurfaceIndex = []
        self.surfaceImpedance = []
        
        # get kwargs
        self.boundary_conditions = None
        self.direction = False
        self.tol = None
        self.parse_input()
        
        # initialize pressures and velocities arrays
        self.u_mesh = np.empty([len(frequency), self.Ns], dtype=object)  # separate sources
        self.u_mesh_Y = np.empty([len(frequency), self.Ns], dtype=object)
        self.p_mesh = np.empty([len(frequency), self.Ns], dtype=object)  # separate drivers
        self.u_total_mesh = np.empty([len(frequency)], dtype=object)   # summed sources
        self.p_total_mesh = np.empty([len(frequency)], dtype=object)   # summed sources
        self.propag_function = np.empty([len(frequency), self.Ns], dtype=object)
        
        # load simulation grid and mirror mesh if needed
        self.grid_sim = bempp_cl.api.import_grid(self.meshPath)
        self.grid_init = bempp_cl.api.import_grid(self.meshPath)
        self.grid_sim, self.sizeFactor = mirror_mesh(self.grid_init, self.boundary_conditions)
        self.vertices = np.shape(self.grid_sim.vertices)[1]
        
        if self.boundary_conditions is not None:
            self.mirror_system()
        else:
            self.xSystem = self.xSource
        
        
        # define space functions
        self.spaceP   = bempp_cl.api.function_space(self.grid_sim, "DP", 0)
        self.identity = sparse.identity(self.spaceP, self.spaceP, self.spaceP)
        
        
        # Assign vibrometric coefficients to corresponding radiators // Assign surface velocity if not vibration data
        self.admittanceCoeff = getSurfaceAdmittance(self.impedanceSurfaceIndex, self.surfaceImpedance, 
                                                    frequency, self.spaceP, self.c_0, self.rho_0)        
        # driver reference
        self.LEM_enclosures = None
        self.radiator = None
        
        
    def parse_input(self):
        """
        Detects and assign kwargs to BEMOBJ variables.

        Returns
        -------
        None.

        """
        if "boundary_conditions" in self.kwargs:
            self.boundary_conditions = self.kwargs["boundary_conditions"].parameters
            self.initialize_conditions()
        elif "boundary_condition" in self.kwargs: # just in case user forget "s"
            self.boundary_conditions = self.kwargs["boundary_condition"].parameters
            self.initialize_conditions()
        if "direction" in self.kwargs:
            self.direction = self.kwargs["direction"]
        if "vibrometry_points" in self.kwargs:
            self.vibrometry_points = self.kwargs["vibrometry_points"]
        if "tol" in self.kwargs:
            self.tol = self.kwargs["tol"]
        else:
            self.tol = 1e-5
            
    def initialize_conditions(self):
        """
        Assign boundary conditions to the impedanceSurfaceIndex and surfaceImpedance 
        variables.

        Returns
        -------
        None.

        """
        
        for bc in self.boundary_conditions:
            if bc not in ["x", "X", "y", "Y", "z", "Z"]:
                self.impedanceSurfaceIndex.append(self.boundary_conditions[bc]["index"])
                self.surfaceImpedance.append(self.boundary_conditions[bc]["impedance"])
            else:
                pass
     
    
    def mirror_system(self):
        self.xSystem = np.copy(self.xSource)
        for param in self.boundary_conditions:
            if param in ["x", "X"]:
                xmirror = np.copy(self.xSystem)
                xmirror[:, 0] = 2*self.boundary_conditions.parameters[param]["offset"] - xmirror[:, 0]
                self.xSystem = np.concatenate([self.xSystem, xmirror])
            if param in ["y", "Y"]:
                ymirror = np.copy(self.xSystem)
                ymirror[:, 1] = 2*self.boundary_conditions.parameters[param]["offset"] - ymirror[:, 1]
                self.xSystem = np.concatenate([self.xSystem, ymirror])
            if param in ["z", "Z"]:
                zmirror = np.copy(self.xSystem)
                zmirror[:, 2] = 2*self.boundary_conditions.parameters[param]["offset"] - zmirror[:, 2]
                self.xSystem = np.concatenate([self.xSystem, zmirror])
        self.QSystem = np.tile(self.QSource, (self.sizeFactor, 1))
        
    
    def solve(self):
        """
        Compute the Boundary Element Method (BEM) solution for the loudspeaker system.

        This method performs the BEM computations to determine the acoustic pressure distribution on the mesh
        due to the contribution of individual speakers. The total pressure distribution is also computed by summing
        up the contributions of all speakers.

        Returns
        -------
        None

        Notes
        -----
        This function calculates the acoustic pressure distribution on the mesh at various frequencies using the
        Boundary Element Method. It iterates through the frequencies and speakers, calculates the necessary BEM
        operators (double layer, single layer), and uses GMRES solver to solve the BEM equation for the acoustic
        pressure distribution. The results are stored in class attributes for further analysis.

        """
        
        if self.domain == "exterior":
            domain_operator = -1
        elif self.domain == "interior":
            domain_operator = +1
        
        omega = 2 * np.pi * self.frequency
        k = -omega / self.c_0
        
        print("Computing pressure on mesh")
        if self.admittanceCoeff is None:
            for i in tqdm(range(len(k))):
                double_layer = helmholtz.double_layer(self.spaceP, self.spaceP,
                                                      self.spaceP, k[i])
                single_layer = helmholtz.single_layer(self.spaceP, self.spaceP,
                                                      self.spaceP, k[i])
                lhs = double_layer + domain_operator*(0.5 * self.identity)
                
                for rs in range(self.Ns):
                    xsource = self.xSystem[rs::self.Ns, :]
                    G  = greenMonopole(xsource, self.grid_sim, k[i])
                    n0 = incidenceCoeff(xsource, self.grid_sim, self.domain)
                    self.propag_function[i, rs] = buildGridFunction(self.spaceP, 
                                                        -1j*k[i]*n0*G)
                    rhs = single_layer * self.propag_function[i, rs]
                    self.u_mesh[i, rs], _ = gmres(lhs, rhs, tol=self.tol)
            self.isComputed = True
            
        elif self.admittanceCoeff is not None:
            for i in tqdm(range(len(k))):
                double_layer = helmholtz.double_layer(self.spaceP, self.spaceP,
                                                      self.spaceP, k[i])
                single_layer = helmholtz.single_layer(self.spaceP, self.spaceP,
                                                      self.spaceP, k[i])

                # ABSORBING SURFACES
                Yn = self.admittanceCoeff[:, i]  # all admittance coeff at current frequency
                yn_fun = bempp_cl.api.GridFunction(self.spaceP, coefficients=Yn)  # ? doubts on its usefulness
                yn = DiagonalOperator(yn_fun.coefficients)
                
                lhs = ((double_layer + 
                       domain_operator*(0.5 * self.identity)).weak_form() - 
                       (1j*k[i]*single_layer.weak_form() * yn))
                
                for rs in range(self.Ns):
                    # PROPAGATION FROM SOURCE TO BOUNDARIES
                    xsource = self.xSystem[rs::self.Ns, :]
                    G  = greenMonopole(xsource, self.grid_sim, k[i])
                    n0 = incidenceCoeff(xsource, self.grid_sim, self.domain)
                    self.propag_function[i, rs] = buildGridFunction(self.spaceP, 
                                                        -1j*k[i]*n0*G)
                    
                    
                    rhs = (single_layer * self.propag_function[i, rs])
                    rhs = rhs.projections(self.spaceP)
                    
                    u_total_coeff, _ = scipy_gmres(lhs, rhs, rtol=self.tol)
                    u_total_coeff_Y = 1j*k[i]*u_total_coeff*yn_fun.coefficients
                    self.u_mesh[i, rs] = bempp_cl.api.GridFunction(self.spaceP, 
                                                                coefficients=u_total_coeff)
                    self.u_mesh_Y[i, rs] = bempp_cl.api.GridFunction(self.spaceP, 
                                                                coefficients=u_total_coeff_Y)

                    
            self.isComputed = True
        self.p_mesh = element2vertice_pressure(self.grid_sim, self.u_mesh)


    def getMicPressure(self, micPosition, individualSpeakers=False):
        """
        Get the pressure received at the considered microphones.

        Parameters
        ----------
        micPosition : numpy array
            Coordinates of the microphones (Cartesian). Shape: (nMic, 3)
        individualSpeakers : bool, optional
            If True, returns an array containing pressure received at each microphone from individual speakers.
            If False, returns the summed pressure received at each microphone from all speakers.
            Default is False.

        Returns
        -------
        pressure_mic : numpy array
            Pressure received at the specified microphones. Shape: (nFreq, nMic)

        Notes
        -----
        This function calculates the acoustic pressure received at the specified microphone positions for each frequency
        in the frequency array. The pressure is computed based on the BEM solution obtained from `computeBEM` method.
        It uses BEM operators (double layer, single layer) and their interactions with the speaker distributions to
        compute the pressure at each microphone position.

        """
        micPosition = np.array(micPosition).T
        nMic = np.shape(micPosition)[1]

        pressure_mic_array = np.zeros([len(self.frequency), nMic, self.Ns], dtype=complex)
        pressure_mic = np.zeros([len(self.frequency), nMic], dtype=complex)
        omega = 2 * np.pi * self.frequency
        k = -omega / self.c_0
        
        
        if self.admittanceCoeff is None:
            for i in tqdm(range(len(k))):
                for rs in range(self.Ns):
                    xsource = self.xSystem[rs::self.Ns, :]
                    single_pot = helmholtz_potential.single_layer(self.spaceP, 
                                                                  micPosition, k[i])
                    double_pot = helmholtz_potential.double_layer(self.spaceP, 
                                                                  micPosition, k[i])
                    for imag in range(len(xsource)):
                        dist = ((xsource[imag, 0]-micPosition[0, :])**2 + 
                                (xsource[imag, 1]-micPosition[1, :])**2 + 
                                (xsource[imag, 2]-micPosition[2, :])**2)**0.5
                        pressure_mic_array[i, :, rs] += np.reshape(np.exp(1j*k[i]*dist) / (4*np.pi*dist)
                             + double_pot.evaluate(self.u_mesh[i, rs])
                             - single_pot.evaluate(self.propag_function[i, rs]), nMic)
                    pressure_mic[i, :] += pressure_mic_array[i, :, rs]
            
            if individualSpeakers is True:
                out = (pressure_mic, pressure_mic_array)
            elif individualSpeakers is False:
                out = pressure_mic
        elif self.admittanceCoeff is not None:
            for i in tqdm(range(len(k))):
                for rs in range(self.Ns):
                    xsource = self.xSystem[rs::self.Ns, :]
                    single_pot = helmholtz_potential.single_layer(self.spaceP, 
                                                                  micPosition, k[i])
                    double_pot = helmholtz_potential.double_layer(self.spaceP, 
                                                                  micPosition, k[i])
                    for imag in range(len(xsource)):
                        dist = ((xsource[imag, 0]-micPosition[0, :])**2 + 
                                (xsource[imag, 1]-micPosition[1, :])**2 + 
                                (xsource[imag, 2]-micPosition[2, :])**2)**0.5
                        pressure_mic_array[i, :, rs] += np.reshape(np.exp(1j*k[i]*dist) / (4*np.pi*dist)
                             + double_pot.evaluate(self.u_mesh[i, rs])
                             - single_pot.evaluate(self.propag_function[i, rs])
                             + single_pot.evaluate(self.u_mesh_Y[i, rs]), nMic)
                    pressure_mic[i, :] += pressure_mic_array[i, :, rs]
            
            if individualSpeakers is True:
                out = (pressure_mic, pressure_mic_array)
            elif individualSpeakers is False:
                out = pressure_mic
        return out


#%% helper function
# def getSurfaceAdmittance(absorbingSurface, surfaceImpedance, freq, spaceP, c_0, rho_0):
#     """
#     Compute the total single layer coefficients linked to surfaces impedance.
#     :param absorbingSurface:
#     :param surfaceImpedance:
#     :param freq:
#     :param spaceP:
#     :return:
#     """
#     Nfft       = len(freq)
#     absSurf_in = np.array(absorbingSurface)
#     surfImp_in = surfaceImpedance
#     Nsurf      = len(absSurf_in)             # number of absorbing surfaces
#     grid       = spaceP.grid
#     dofCount   = spaceP.grid_dof_count

#     if absSurf_in.shape[0] == 0 :
#         admittanceMatrix = None
#     else:
#         admittanceMatrix = np.ones([dofCount, Nfft], dtype=complex) * 2.5e-3 # corresponds to 0.5% damping
#         for surf in range(Nsurf):
#             tmp_surf = absSurf_in[surf]  # current surface on which we apply admittance coefficients
#             vertex, _ = get_group_points(grid, tmp_surf)
#             for f in range(Nfft):
#                 try:
#                     Yn = rho_0 * c_0 / surfImp_in[surf][f] # rho_0 * c_0
#                 except:
#                     Yn = rho_0 * c_0 / surfImp_in[surf] # rho_0 * c_0
#                 admittanceMatrix[vertex, f] = np.ones(len(vertex)) * Yn
#     return admittanceMatrix      


"""Maxwell potential operators."""

import numpy as _np


def electric_field(
    space,
    points,
    wavenumber,
    parameters=None,
    assembler="dense",
    device_interface=None,
    precision=None,
):
    """Return a Maxwell electric field potential operator."""
    from bempp_cl.api.operators import OperatorDescriptor
    from bempp_cl.api.assembly.potential_operator import PotentialOperator
    from bempp_cl.api.assembly.assembler import PotentialAssembler
    import bempp_cl.api

    if space.identifier != "rwg0":
        raise ValueError("Space must be an RWG type function space.")

    if precision is None:
        precision = bempp_cl.api.DEFAULT_PRECISION

    operator_descriptor = OperatorDescriptor(
        "maxwell_electric_field_potential",  # Identifier
        [_np.real(wavenumber), _np.imag(wavenumber)],  # Options
        "helmholtz_single_layer",  # Kernel type
        "maxwell_electric_field",  # Assembly type
        precision,  # Precision
        True,  # Is complex
        None,  # Singular part
        3,  # Kernel dimension
    )

    return PotentialOperator(
        PotentialAssembler(space, points, operator_descriptor, device_interface, assembler, parameters)
    )


def magnetic_field(
    space,
    points,
    wavenumber,
    parameters=None,
    assembler="dense",
    device_interface=None,
    precision=None,
):
    """Return a Maxwell magnetic field potential operator."""
    from bempp_cl.api.operators import OperatorDescriptor
    from bempp_cl.api.assembly.potential_operator import PotentialOperator
    from bempp_cl.api.assembly.assembler import PotentialAssembler
    import bempp_cl.api

    if space.identifier != "rwg0":
        raise ValueError("Space must be an RWG type function space.")

    if precision is None:
        precision = bempp_cl.api.DEFAULT_PRECISION

    operator_descriptor = OperatorDescriptor(
        "maxwell_magnetic_field_potential",  # Identifier
        [_np.real(wavenumber), _np.imag(wavenumber)],  # Options
        "helmholtz_single_layer",  # Kernel type
        "maxwell_magnetic_field",  # Assembly type
        precision,  # Precision
        True,  # Is complex
        None,  # Singular part
        3,  # Kernel dimension
    )

    return PotentialOperator(
        PotentialAssembler(space, points, operator_descriptor, device_interface, assembler, parameters)
    )



"""OpenCL routines."""

import pyopencl as _cl
import os as _os

_CURRENT_PATH = _os.path.dirname(_os.path.realpath(__file__))
_INCLUDE_PATH = _os.path.abspath(_os.path.join(_CURRENT_PATH, "./sources/include"))
_KERNEL_PATH = _os.path.abspath(_os.path.join(_CURRENT_PATH, "./sources/kernels"))

_DEFAULT_CPU_DEVICE = None
_DEFAULT_CPU_CONTEXT = None

_DEFAULT_GPU_DEVICE = None
_DEFAULT_GPU_CONTEXT = None


def select_cl_kernel(operator_descriptor, mode):
    """Select OpenCL kernel."""
    singular_assemblers = {
        "default_scalar": "evaluate_dense_singular",
        "laplace_hypersingular": "evaluate_dense_laplace_hypersingular_singular",
        "helmholtz_hypersingular": "evaluate_dense_helmholtz_hypersingular_singular",
        "modified_helmholtz_hypersingular": "evaluate_dense_helmholtz_hypersingular_singular",
        "maxwell_electric_field": "evaluate_dense_electric_field_singular",
        "maxwell_magnetic_field": "evaluate_dense_magnetic_field_singular",
    }

    regular_assemblers = {
        "default_scalar": "evaluate_dense_regular",
        "laplace_hypersingular": "evaluate_dense_laplace_hypersingular_regular",
        "helmholtz_hypersingular": "evaluate_dense_helmholtz_hypersingular_regular",
        "modified_helmholtz_hypersingular": "evaluate_dense_helmholtz_hypersingular_regular",
        "maxwell_electric_field": "evaluate_dense_electric_field_regular",
        "maxwell_magnetic_field": "evaluate_dense_magnetic_field_regular",
    }

    potential_assemblers = {
        "default_scalar": "evaluate_scalar_potential",
        "maxwell_electric_field": "evaluate_electric_field_potential",
        "maxwell_magnetic_field": "evaluate_magnetic_field_potential",
        "maxwell_electric_far_field": "evaluate_maxwell_electric_far_field",
        "maxwell_magnetic_far_field": "evaluate_maxwell_magnetic_far_field",
    }

    kernels = {
        "laplace_single_layer": "laplace_single_layer",
        "laplace_double_layer": "laplace_double_layer",
        "laplace_adjoint_double_layer": "laplace_adjoint_double_layer",
        "helmholtz_single_layer": "helmholtz_single_layer",
        "helmholtz_double_layer": "helmholtz_double_layer",
        "helmholtz_far_field_single_layer": "helmholtz_single_layer_far_field",
        "helmholtz_far_field_double_layer": "helmholtz_double_layer_far_field",
        "helmholtz_adjoint_double_layer": "helmholtz_adjoint_double_layer",
        "modified_helmholtz_single_layer": "modified_helmholtz_real_single_layer",
        "modified_helmholtz_double_layer": "modified_helmholtz_real_double_layer",
        "modified_helmholtz_adjoint_double_layer": "modified_helmholtz_real_adjoint_double_layer",
    }

    if mode == "singular":
        return (
            singular_assemblers[operator_descriptor.assembly_type],
            kernels[operator_descriptor.kernel_type],
        )

    elif mode == "regular":
        return (
            regular_assemblers[operator_descriptor.assembly_type],
            kernels[operator_descriptor.kernel_type],
        )
    elif mode == "potential":
        return (
            potential_assemblers[operator_descriptor.assembly_type],
            kernels[operator_descriptor.kernel_type],
        )
    else:
        raise ValueError(f"Unknown mode {mode}")


def get_kernel_compile_options(options, precision):
    """Create compiler options from parameters and precision."""
    import numbers

    if precision == "single":
        literal = "f"
    else:
        literal = ""

    compile_options = []
    for key, value in options.items():
        if value is None:
            compile_options += ["-D", "{0}".format(key)]
        else:
            if isinstance(value, numbers.Real) and not isinstance(value, numbers.Integral):
                value_string = str(value) + literal
            else:
                value_string = str(value)
            compile_options += ["-D", "{0}={1}".format(key, value_string)]

    compile_options += ["-I", _INCLUDE_PATH]

    # Add precision flag

    if precision == "single":
        val = 0
    else:
        val = 1

    compile_options.append("-DPRECISION={0}".format(val))

    return compile_options


def build_program(assembly_function, options, precision, device_type="cpu"):
    """Build the kernel and return it."""
    file_name = assembly_function + ".cl"
    kernel_file = _os.path.join(_KERNEL_PATH, file_name)

    kernel_file_tmp = open(kernel_file)
    kernel_string = kernel_file_tmp.read() #open(kernel_file).read()
    kernel_options = get_kernel_compile_options(options, precision)
    
    kernel_file_tmp.close()
   
    return _cl.Program(default_context(device_type), kernel_string).build(options=kernel_options).kernel_function


def get_kernel_from_operator_descriptor(operator_descriptor, options, mode, force_novec=False, device_type="cpu"):
    """Return compiled kernel from operator descriptor."""
    precision = operator_descriptor.precision
    assembly_function, kernel_name = select_cl_kernel(operator_descriptor, mode=mode)

    vec_length = get_vector_width(precision, device_type)
    vec_string = get_vec_string(precision, device_type)

    if not mode == "singular":
        if force_novec or vec_length == 1:
            assembly_function += "_novec"
        else:
            assembly_function += "_vec"
    options["KERNEL_FUNCTION"] = kernel_name
    options["VEC_LENGTH"] = vec_length
    options["VEC_STRING"] = vec_string
    return build_program(assembly_function, options, precision, device_type)


def get_kernel_from_name(name, options, precision="double", device_type="cpu"):
    """Return compiled kernel from name."""

    vec_length = get_vector_width(precision, device_type)
    vec_string = get_vec_string(precision, device_type)

    options["VEC_LENGTH"] = vec_length
    options["VEC_STRING"] = vec_string
    return build_program(name, options, precision, device_type)


def get_vector_width(precision, device_type="cpu"):
    """Return vector width."""
    import bempp_cl.api

    mode_to_length = {"novec": 1, "vec4": 4, "vec8": 8, "vec16": 16}

    if device_type == "gpu":
        return 1
    if bempp_cl.api.VECTORIZATION_MODE == "auto":
        return get_native_vector_width(default_device(device_type), precision)
    else:
        return mode_to_length[bempp_cl.api.VECTORIZATION_MODE]


def get_vec_string(precision, device_type="cpu"):
    """Return vectorisation string."""
    vec_strings = {1: "novec", 4: "vec4", 8: "vec8", 16: "vec16"}

    return vec_strings[get_vector_width(precision, device_type)]


def default_cpu_device():
    """Return the default CPU device."""
    import bempp_cl.api
    import os

    # pylint: disable=W0603
    global _DEFAULT_CPU_DEVICE
    global _DEFAULT_CPU_CONTEXT

    if "BEMPP_CPU_DRIVER" in os.environ:
        name = os.environ["BEMPP_CPU_DRIVER"]
    else:
        name = None

    if _DEFAULT_CPU_DEVICE is None:
        try:
            ctx, device = find_cpu_driver(name)
        except:  # noqa: E722
            raise RuntimeError("Could not find suitable OpenCL CPU driver.")
        _DEFAULT_CPU_CONTEXT = ctx
        _DEFAULT_CPU_DEVICE = device
        bempp_cl.api.log(f"OpenCL CPU Device set to: {_DEFAULT_CPU_DEVICE.name}")
    return _DEFAULT_CPU_DEVICE


def default_gpu_device():
    """Return the default GPU device."""
    import bempp_cl.api
    import os

    # pylint: disable=W0603
    global _DEFAULT_GPU_DEVICE
    global _DEFAULT_GPU_CONTEXT

    if "BEMPP_GPU_DRIVER" in os.environ:
        name = os.environ["BEMPP_GPU_DRIVER"]
    else:
        name = None

    if _DEFAULT_GPU_DEVICE is None:
        try:
            ctx, device = find_gpu_driver(name)
        except:  # noqa: E722
            raise RuntimeError("Could not find a suitable OpenCL GPU driver.")
        _DEFAULT_GPU_CONTEXT = ctx
        _DEFAULT_GPU_DEVICE = device
        bempp_cl.api.log(f"OpenCL GPU Device set to: {_DEFAULT_GPU_DEVICE.name}")
    return _DEFAULT_GPU_DEVICE


def default_device(device_type="cpu"):
    """Return default device."""

    if device_type == "cpu":
        return default_cpu_device()
    elif device_type == "gpu":
        return default_gpu_device()
    else:
        raise ValueError(f"Unknown value for 'mode' = {device_type}.")


def default_context(device_type="cpu"):
    """Return the default context"""

    if device_type == "cpu":
        return default_cpu_context()
    elif device_type == "gpu":
        return default_gpu_context()
    else:
        raise ValueError(f"Unknown value for 'mode' = {device_type}.")


def default_cpu_context():
    """Return default CPU context."""

    if _DEFAULT_CPU_CONTEXT is None:
        default_cpu_device()

    return _DEFAULT_CPU_CONTEXT


def default_gpu_context():
    """Return default GPU context."""

    if _DEFAULT_GPU_CONTEXT is None:
        default_gpu_device()

    return _DEFAULT_GPU_CONTEXT


def find_cpu_driver(name=None):
    """Find the first available CPU OpenCL driver."""
    found = False
    ctx = None
    device = None
    for platform in _cl.get_platforms():
        if found:
            break
        if name and name not in platform.name:
            continue
        ctx = _cl.Context(
            dev_type=_cl.device_type.ALL,
            properties=[(_cl.context_properties.PLATFORM, platform)],
        )
        for device in ctx.devices:
            if device.type == _cl.device_type.CPU:
                found = True
    if not found:
        raise ValueError(f"Could not find CPU driver containing name {name}.")
    return ctx, device


def find_gpu_driver(name=None):
    """Find the first available GPU OpenCL driver."""

    found = False
    ctx = None
    device = None
    for platform in _cl.get_platforms():
        if found:
            break
        if name and name not in platform.name:
            continue
        ctx = _cl.Context(
            dev_type=_cl.device_type.ALL,
            properties=[(_cl.context_properties.PLATFORM, platform)],
        )
        for device in ctx.devices:
            if device.type == _cl.device_type.GPU:
                found = True
    if not found:
        raise ValueError(f"Could not find GPU driver containing name {name}.")
    return ctx, device


def set_default_cpu_device_by_name(name):
    """
    Set default CPU device by name.

    This method looks for the given string in the available OpenCL
    drivers and picks the first one that contains the given search
    string.

    """
    import bempp_cl.api

    global _DEFAULT_CPU_CONTEXT
    global _DEFAULT_CPU_DEVICE

    try:
        context, device = find_cpu_driver(name)
    except:  # noqa: E722
        raise RuntimeError("No CPU driver with given name found.")

    _DEFAULT_CPU_CONTEXT = context
    _DEFAULT_CPU_DEVICE = device
    vector_width_single = _DEFAULT_CPU_DEVICE.native_vector_width_float
    vector_width_double = _DEFAULT_CPU_DEVICE.native_vector_width_double

    bempp_cl.api.log(
        f"Default CPU device: {_DEFAULT_CPU_DEVICE.name}. "
        + f"Native vector width: {vector_width_single} (single) / "
        + f"{vector_width_double} (double)."
    )


def set_default_cpu_device(platform_index, device_index):
    """Set the default CPU device."""
    import bempp_cl.api

    # pylint: disable=W0603
    global _DEFAULT_CPU_DEVICE
    global _DEFAULT_CPU_CONTEXT

    platform = _cl.get_platforms()[platform_index]
    device = platform.get_devices()[device_index]
    _DEFAULT_CPU_CONTEXT = _cl.Context(devices=[device], properties=[(_cl.context_properties.PLATFORM, platform)])
    _DEFAULT_CPU_DEVICE = _DEFAULT_CPU_CONTEXT.devices[0]

    vector_width_single = _DEFAULT_CPU_DEVICE.native_vector_width_float
    vector_width_double = _DEFAULT_CPU_DEVICE.native_vector_width_double

    bempp_cl.api.log(
        f"Default CPU device: {_DEFAULT_CPU_DEVICE.name}. "
        + f"Native vector width: {vector_width_single} (single) / "
        + f"{vector_width_double} (double)."
    )


def set_default_gpu_device_by_name(name):
    """
    Set default GPU device by name.

    This method looks for the given string in the available OpenCL
    drivers and picks the first one that contains the given search
    string.

    """
    import bempp_cl.api

    global _DEFAULT_GPU_CONTEXT
    global _DEFAULT_GPU_DEVICE

    try:
        pair = find_gpu_driver(name)
        context, device = pair[0], pair[1]
    except:  # noqa: E722
        raise RuntimeError("No GPU driver with given name found.")

    _DEFAULT_GPU_CONTEXT = context
    _DEFAULT_GPU_DEVICE = device
    vector_width_single = _DEFAULT_GPU_DEVICE.native_vector_width_float
    vector_width_double = _DEFAULT_GPU_DEVICE.native_vector_width_double

    bempp_cl.api.log(
        f"Default GPU device: {_DEFAULT_GPU_DEVICE.name}. "
        + f"Native vector width: {vector_width_single} (single) / "
        + f"{vector_width_double} (double)."
    )


def set_default_gpu_device(platform_index, device_index):
    """Set the default GPU device."""
    import bempp_cl.api

    # pylint: disable=W0603
    global _DEFAULT_GPU_DEVICE
    global _DEFAULT_GPU_CONTEXT

    platform = _cl.get_platforms()[platform_index]
    device = platform.get_devices()[device_index]
    _DEFAULT_GPU_CONTEXT = _cl.Context(devices=[device], properties=[(_cl.context_properties.PLATFORM, platform)])
    _DEFAULT_GPU_DEVICE = _DEFAULT_GPU_CONTEXT.devices[0]

    vector_width_single = _DEFAULT_GPU_DEVICE.native_vector_width_float
    vector_width_double = _DEFAULT_GPU_DEVICE.native_vector_width_double

    bempp_cl.api.log(
        f"Default GPU device: {_DEFAULT_GPU_DEVICE.name}. "
        + f"Native vector width: {vector_width_single} (single) / "
        + f"{vector_width_double} (double)."
    )


def show_available_platforms_and_devices():
    """Print available platforms and devices."""
    platforms = _cl.get_platforms()
    for platform_index, platform in enumerate(platforms):
        print(str(platform_index) + ": " + platform.get_info(_cl.platform_info.NAME))
        devices = platform.get_devices()
        for device_index, device in enumerate(devices):
            print(4 * " " + str(device_index) + ": " + device.get_info(_cl.device_info.NAME))


def get_native_vector_width(device, precision):
    """Get default vector width for device."""
    if precision == "single":
        return device.native_vector_width_float
    elif precision == "double":
        return device.native_vector_width_double
    else:
        raise ValueError("precision must be one of 'single', 'double'.")



"""Definition of potential operators."""


class PotentialOperator(object):
    """Provides an interface to potential operators.

    This class is not supposed to be instantiated directly.
    """

    def __init__(self, potential_evaluator):
        """Construct. Should not be called by the user."""
        self._evaluator = potential_evaluator

    def evaluate(self, grid_fun):
        """
        Apply the potential operator to a grid function.

        Parameters
        ----------
        grid_fun : bempp_cl.api.GridFunction
            A GridFunction object that represents the boundary density to
            which the potential is applied to.

        """
        return self._evaluator.evaluate(grid_fun.coefficients)

    def _is_compatible(self, other):
        """Check compatibility with other potential operator."""
        import numpy as np

        return (
            self.component_count == other.component_count
            and np.linalg.norm(self.evaluation_points - other.evaluation_points, ord=np.inf) == 0
            and self.space.is_compatible(other.space)
        )

    def __add__(self, obj):
        """Add."""
        if not self._is_compatible(obj):
            raise ValueError("Potential operators not compatible.")

        return _SumPotentialOperator(self, obj)

    def __mul__(self, obj):
        """Multiply."""
        import numpy as np
        from bempp_cl.api import GridFunction

        if np.isscalar(obj):
            return _ScaledPotentialOperator(self, obj)
        elif isinstance(obj, GridFunction):
            return self.evaluate(obj)
        else:
            return NotImplemented

    def __matmul__(self, obj):
        """Multiply."""
        return self.__mul__(obj)

    def __rmul__(self, obj):
        """Reverse multiply."""
        import numpy as np

        if np.isscalar(obj):
            return _ScaledPotentialOperator(self, obj)
        else:
            return NotImplemented

    def __neg__(self):
        """Negate."""
        return self.__mul__(-1.0)

    def __sub__(self, other):
        """Subtract."""
        return self.__add__(-other)

    @property
    def space(self):
        """Return the underlying function space."""
        return self._evaluator.space

    @property
    def component_count(self):
        """Return number of components of the potential (1 for scalar potentials)."""
        return self._evaluator.kernel_dimension

    @property
    def evaluation_points(self):
        """Return the evaluation points."""
        return self._evaluator.points


class _ScaledPotentialOperator(PotentialOperator):
    """Scaled potential operator."""

    def __init__(self, op, alpha):
        self._op = op
        self._alpha = alpha

    def evaluate(self, grid_fun):
        """
        Apply the potential operator to a grid function.

        Parameters
        ----------
        grid_fun : bempp_cl.api.GridFunction
            A GridFunction object that represents the boundary density to
            which the potential is applied to.

        """
        return self._alpha * self._op.evaluate(grid_fun)

    @property
    def space(self):
        """Return the underlying function space."""
        return self._op.space

    @property
    def component_count(self):
        """Return number of components of the potential (1 for scalar potentials)."""
        return self._op.component_count

    @property
    def evaluation_points(self):
        """Return the evaluation points."""
        return self._op.points


class _SumPotentialOperator(PotentialOperator):
    """Sum of two potential operators."""

    def __init__(self, op1, op2):
        """Create sum of two potential operators."""
        if not op1._is__compatible(op2):
            raise ValueError("Potential operators are not compatible.")

        self._op1 = op1
        self._op2 = op2

    def evaluate(self, grid_fun):
        """
        Apply the potential operator to a grid function.

        Parameters
        ----------
        grid_fun : bempp_cl.api.GridFunction
            A GridFunction object that represents the boundary density to
            which the potential is applied to.

        """
        return self._op1.evaluate(grid_fun) + self._op2.evaluate(grid_fun)

    @property
    def space(self):
        """Return the underlying function space."""
        return self._op1.space

    @property
    def component_count(self):
        """Return number of components of the potential (1 for scalar potentials)."""
        return self._op1.component_count

    @property
    def evaluation_points(self):
        """Return the evaluation points."""
        return self._op1.points



"""Routines for import and export."""

import meshio as _meshio
import numpy as _np


def import_grid(filename):
    """
    Import a grid.

    This routine uses the meshio library to export grids.
    A number of types are supported, including vtk, vtu,
    gmsh, dolphin xml. For a full list see

    https://github.com/nschloe/meshio

    """
    from bempp_cl.api.grid.grid import Grid

    mesh = _meshio.read(filename)

    vertices = mesh.points.T
    elements = mesh.cells_dict["triangle"].T.astype("uint32")

    try:
        domain_indices = mesh.cell_data_dict["gmsh:physical"]["triangle"]
    except:  # noqa: E722
        domain_indices = None

    if domain_indices is None or _np.all(domain_indices == 0):
        try:
            domain_indices = mesh.cell_data_dict["gmsh:geometrical"]["triangle"]
        except:  # noqa: E722
            pass

    return Grid(vertices, elements, domain_indices=domain_indices)


def export(
    filename,
    grid=None,
    grid_function=None,
    data_type=None,
    transformation=None,
    write_binary=True,
):
    """
    Export grids and grid functions.

    This method internally uses the meshio library. For a full
    list of supported data types see

    https://github.com/nschloe/meshio

    Note that export of domain indices is only possible for Gmsh (.msh)
    format files.

    Parameters
    ----------
    filename : string
        The name of the file to write out. The data type is
        chosen based on the file ending.
    grid : Grid object
        A grid object to export.
    grid_function : GridFunction object
        Grid function to export
    data_type : string
        Either 'node' for vertex data or 'element' for data
        at element centers.
    transformation : string or callable
        One of 'real', 'imag', 'abs', 'log_abs',
        None or a callable object. Transforms the
        data on input. A callable must return numpy
        arrays with the same number of dimensions as
        the input. If transformation is None the data
        is not modified.
    write_binary : Boolean
        Use binary format (write_binary=True) for the
        data if supported by the file format.
    """
    import os

    if data_type is None and grid_function is not None:
        if grid_function.space.identifier == "p1":
            data_type = "node"
        else:
            data_type = "element"

    _, extension = os.path.splitext(filename)

    file_format = None

    if extension == ".msh":
        # Ensure that we use Gmsh2 for output
        # to preserve domain indices.
        # meshio does not yet support domain
        # indices for gmsh4.
        gmsh = True
        if write_binary:
            file_format = "gmsh22"
        else:
            file_format = "gmsh22"
    else:
        gmsh = False

    if grid is not None and grid_function is not None:
        raise ValueError("Exactly one of 'grid' and 'grid_function' must be supplied.")

    cell_data = {}
    point_data = None

    if grid_function is not None:
        grid = grid_function.space.grid

        if data_type == "node":
            data = _transform_array(grid_function.evaluate_on_vertices(), transformation).T
            if _np.iscomplexobj(data):
                point_data = {"real": _np.real(data), "imag": _np.imag(data)}
            else:
                point_data = {"data": data}
        elif data_type == "element":
            data = _transform_array(grid_function.evaluate_on_element_centers(), transformation).T
            if _np.iscomplexobj(data):
                cell_data["real"] = _np.real(data)
                cell_data["imag"] = _np.imag(data)
            else:
                cell_data["data"] = _np.array([data])

        else:
            raise ValueError("'data_type' must be one of 'element' or 'node'")

    cells = [("triangle", grid.elements.T.astype("int32"))]
    points = grid.vertices.T

    if gmsh:
        # physical and geometrical index must be 2 dim arrays with first dimension
        # equal to the number of different element types (here always 1
        cell_data["gmsh:physical"] = grid.domain_indices.astype("int32").reshape((1, -1))
        unique_dom_indices = set(grid.domain_indices)
        unique_geom_indices = range(1, 1 + len(unique_dom_indices))
        geom_indices_map = dict(zip(unique_dom_indices, unique_geom_indices))
        geom_indices = _np.array(
            [geom_indices_map[dom_index] for dom_index in grid.domain_indices],
            dtype="int32",
        )
        cell_data["gmsh:geometrical"] = geom_indices.reshape((1, -1))
    else:
        cell_data["domain_index"] = grid.domain_indices.astype("int32").reshape((1, -1))

    _meshio.write_points_cells(
        filename,
        points,
        cells,
        point_data=point_data,
        cell_data=cell_data,
        file_format=file_format,
        binary=write_binary,
    )


def _transform_array(a, mode=None):
    """
    Transform a data array.

    Parameters
    ----------
    a : np.ndarray
        Either a scalar array or a two dimensional data array.
    mode : string, callable or None
        One of 'real', 'imag', 'abs', 'log_abs', 'abs_squared',
        a transformation callable or None. The callable needs to take
        an input array of dimension 2 and return an array of
        dimension 2. If mode is None the input array is not modified.
    """
    if mode is None:
        return a

    ndim = a.ndim
    if ndim == 1:
        _np.expand_dims(a, 1)

    if mode == "real":
        res = _np.real(a)
    elif mode == "imag":
        res = _np.imag(a)
    elif mode == "abs":
        res = _np.sqrt(_np.sum(_np.abs(a) ** 2, axis=0, keepdims=True))
    elif mode == "abs_squared":
        res = _np.sum(_np.abs(a) ** 2, axis=0, keepdims=True)
    elif mode == "log_abs":
        res = _np.log(_np.sqrt(_np.sum(_np.abs(a) ** 2, axis=0, keepdims=True)))
    else:
        res = mode(a)

    if ndim == 1:
        return _np.squeeze(res, axis=0)
    else:
        return res



"""Dispatch kernel calls to different implementations."""


def singular_assembler_dispatcher(device_interface, *args):
    """Dispatch the singular assembler to the different implementations."""
    interface_type = device_interface.split("_")[0]

    if interface_type == "opencl":
        from bempp_cl.core.opencl_assemblers import singular_assembler

        singular_assembler(device_interface, *args)

    elif interface_type == "numba":
        from bempp_cl.core.numba_assemblers import singular_assembler

        singular_assembler(device_interface, *args)

    else:
        raise ValueError("Device interface must be one of 'numba', 'opencl'.")


def dense_assembler_dispatcher(device_interface, *args):
    """Dispatcher for dense assemblers."""
    interface_type = device_interface.split("_")[0]

    if interface_type == "opencl":
        from bempp_cl.core.opencl_assemblers import dense_assembler

        dense_assembler(device_interface, *args)

    elif interface_type == "numba":
        from bempp_cl.core.numba_assemblers import dense_assembler

        dense_assembler(device_interface, *args)

    else:
        raise ValueError("Device interface must be one of 'numba', 'opencl'.")


def potential_dispatcher(device_interface, *args):
    """Potential assembler dispatcher."""
    interface_type = device_interface.split("_")[0]

    if interface_type == "numba":
        from bempp_cl.core.numba_assemblers import potential_assembler

        return potential_assembler(device_interface, *args)

    elif interface_type == "opencl":
        from bempp_cl.core.opencl_assemblers import potential_assembler

        return potential_assembler(device_interface, *args)

    else:
        raise ValueError("Device interface must be one of 'numba', 'opencl'.")



"""Interfaces to modified Helmholtz operators."""

from bempp_cl.api.operators.boundary import common as _common
import numpy as _np


def single_layer(
    domain,
    range_,
    dual_to_range,
    omega,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the Helmholtz single-layer boundary operator."""
    if _np.imag(omega) != 0:
        raise ValueError("'omega' must be real.")

    return _common.create_operator(
        "modified_helmholtz_single_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [omega],
        "modified_helmholtz_single_layer",
        "default_scalar",
        device_interface,
        precision,
        False,
    )


def double_layer(
    domain,
    range_,
    dual_to_range,
    omega,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the mod. Helmholtz double-layer boundary operator."""
    if _np.imag(omega) != 0:
        raise ValueError("'omega' must be real.")

    return _common.create_operator(
        "modified_helmholtz_double_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [omega],
        "modified_helmholtz_double_layer",
        "default_scalar",
        device_interface,
        precision,
        False,
    )


def adjoint_double_layer(
    domain,
    range_,
    dual_to_range,
    omega,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the mod. Helmholtz adj. double-layer boundary op."""
    if _np.imag(omega) != 0:
        raise ValueError("'omega' must be real.")

    return _common.create_operator(
        "modified_helmholtz_adjoint_double_layer_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [omega],
        "modified_helmholtz_adjoint_double_layer",
        "default_scalar",
        device_interface,
        precision,
        False,
    )


def hypersingular(
    domain,
    range_,
    dual_to_range,
    omega,
    parameters=None,
    assembler="default_nonlocal",
    device_interface=None,
    precision=None,
):
    """Assemble the mod. Helmholtz hypersingular boundary op."""
    if domain.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Domain shapeset must be of type 'p1_discontinuous'.")

    if dual_to_range.shapeset.identifier != "p1_discontinuous":
        raise ValueError("Dual to range shapeset must be of type 'p1_discontinuous'.")

    if _np.imag(omega) != 0:
        raise ValueError("'omega' must be real.")

    return _common.create_operator(
        "modified_helmholtz_hypersingular_boundary",
        domain,
        range_,
        dual_to_range,
        parameters,
        assembler,
        [omega],
        "modified_helmholtz_single_layer",
        "modified_helmholtz_hypersingular",
        device_interface,
        precision,
        False,
    )



"""Definition of blocked operator structures."""

import numpy as _np
from bempp_cl.api.assembly.discrete_boundary_operator import _DiscreteOperatorBase


def _sum(op1, op2):
    """Sum of two operators, allowing for one of them to be None."""
    if op1 is None:
        return op2
    elif op2 is None:
        return op1
    else:
        return op1 + op2


def _prod(op1, op2):
    """Product of two operators, allowing for one of them to be None."""
    if op1 is None or op2 is None:
        return None
    else:
        return op1 * op2


class BlockedOperatorBase(object):
    """Base class for blocked operators."""

    def __init__(self):
        """Create base class for blocked operators."""
        self._weak_form = None
        self._range_map = None
        self._cached = None

    def weak_form(self):
        """Return cached weak form (assemble if necessary)."""
        if not self._cached:
            self._cached = self._assemble()
        return self._cached

    def strong_form(self):
        """Return a discrete operator that maps into the range space.

        Parameters
        ----------
        recompute : bool
            Usually the strong form is cached. If this parameter is set to
            `true` the strong form is recomputed.
        """
        from bempp_cl.api.utils.helpers import get_inverse_mass_matrix

        if self._range_map is None:
            nrows = len(self.range_spaces)

            _range_ops = _np.empty((nrows, nrows), dtype="O")

            for index in range(nrows):
                _range_ops[index, index] = get_inverse_mass_matrix(
                    self.range_spaces[index], self.dual_to_range_spaces[index]
                )

            self._range_map = BlockedDiscreteOperator(_range_ops)

        return self._range_map * self.weak_form()

    @property
    def ndims(self):
        """Return number of block rows and block columns."""
        return (len(self.range_spaces), len(self.domain_spaces))

    @property
    def range_spaces(self):
        """Return a list of range spaces of the blocked operator."""
        raise NotImplementedError()

    @property
    def dual_to_range_spaces(self):
        """Return a list of dual to range spaces of the blocked operator."""
        raise NotImplementedError()

    @property
    def domain_spaces(self):
        """Return a list of domain spaces of the blocked operator."""
        raise NotImplementedError()

    def __add__(self, other):
        """Add two blocked operators."""
        if not isinstance(other, BlockedOperatorBase):
            return NotImplementedError

        return SumBlockedOperator(self, other)

    def __neg__(self):
        """Multiply the blocked operator with -1."""
        return self.__mul__(-1)

    def __sub__(self, other):
        """Subtract another blocked operator from this blocked operator."""
        return self.__add__(-other)

    def __mul__(self, other):
        """Multiply two blocked operators."""
        from collections.abc import Iterable

        if _np.isscalar(other):
            # Multiplication with scalar
            return ScaledBlockedOperator(self, other)
        elif isinstance(other, BlockedOperatorBase):
            # Multiplication with another blocked operator.
            return ProductBlockedOperator(self, other)
        elif isinstance(other, Iterable):
            # Multiplication with a list of grid functions.
            from bempp_cl.api.assembly.grid_function import GridFunction

            list_input = list(other)
            if len(list_input) != self.ndims[1]:
                raise ValueError(
                    "Length of input list is {0}.".format(len(list_input))
                    + ". But domain dimension of blocked operator"
                    + "is {0}".format(self.ndims[1])
                )
            for item in list_input:
                if not isinstance(item, GridFunction):
                    raise ValueError("All items in the input list must be grid functions.")
            weak_op = self.weak_form()
            x_in = coefficients_from_grid_functions_list(list_input)
            res = weak_op * x_in

            # Now assemble the output grid functions back together.
            output_list = grid_function_list_from_projections(res, self.range_spaces, self.dual_to_range_spaces)
            return output_list

        else:
            return NotImplemented

    def __matmul__(self, other):
        """Multiply two operators."""
        return self.__mul__(other)

    def __rmul__(self, other):
        """Multiply two operators."""
        if _np.isscalar(other):
            return self.__mul__(other)
        else:
            return NotImplemented


class BlockedOperator(BlockedOperatorBase):
    """Construct an m x n blocked boundary operator."""

    def __init__(self, m, n):
        """Construct an (m x n) blocked boundary operator."""
        super().__init__()

        self._operators = _np.empty((m, n), dtype=object)
        self._rows = m * [False]
        self._cols = n * [False]

        self._dual_to_range_spaces = m * [None]
        self._range_spaces = m * [None]
        self._domain_spaces = n * [None]

        super().__init__()

    def __getitem__(self, key):
        """Return a component operator with a key of the form (i, j)."""
        import bempp_cl.api

        if self._operators[key] is None:
            return bempp_cl.api.ZeroBoundaryOperator(
                self.domain_spaces[key[1]],
                self.range_spaces[key[0]],
                self.dual_to_range_spaces[key[0]],
            )
        else:
            return self._operators[key]

    def __setitem__(self, key, operator):
        """Set a component operator with a key of the form (i, j)."""
        row = key[0]
        col = key[1]

        if self.range_spaces[row] is not None:
            if operator.range != self.range_spaces[row]:
                raise ValueError("Range space not compatible with " + "self.range_spaces[{0}]".format(row))

        if self.dual_to_range_spaces[row] is not None:
            if operator.dual_to_range != self.dual_to_range_spaces[row]:
                raise ValueError(
                    "Dual to range space not compatible with " + "self.dual_to_range_spaces[{0}]".format(row)
                )

        if self.domain_spaces[col] is not None:
            if operator.domain != self.domain_spaces[col]:
                raise ValueError("Domain space not compatible with " + "self.domain_spaces[{0}]".format(col))

        self._range_spaces[row] = operator.range
        self._dual_to_range_spaces[row] = operator.dual_to_range
        self._domain_spaces[col] = operator.domain
        self._operators[key] = operator
        self._rows[row] = True
        self._cols[col] = True

    def _fill_complete(self):
        """Check if each row and  column contain at least one operator."""
        return (False not in self._cols) and (False not in self._rows)

    def _assemble(self):
        """Implement the weak form."""
        if not self._fill_complete():
            raise ValueError("Each row and column must have at least one operator")

        ops = _np.empty((self.ndims[0], self.ndims[1]), dtype="O")

        for i in range(self.ndims[0]):
            for j in range(self.ndims[1]):
                if self._operators[i, j] is not None:
                    ops[i, j] = self._operators[i, j].weak_form()

        return BlockedDiscreteOperator(ops)

    @property
    def range_spaces(self):
        """Return the list of range spaces."""
        return tuple(self._range_spaces)

    @property
    def dual_to_range_spaces(self):
        """Return the list of dual_to_range spaces."""
        return tuple(self._dual_to_range_spaces)

    @property
    def domain_spaces(self):
        """Return the list of domain spaces."""
        return tuple(self._domain_spaces)


class GeneralizedBlockedOperator(BlockedOperatorBase):
    """
    Construct a generalized blocked operator.

    A generalized blocked operator has as components either

    - Simple operators
    - Blocked operators
    - Generalized blocked operators
    - Arrays of simple/blocked/generalized blocked operators

    """

    def __init__(self, array):
        """
        Initialize the operator.

        The input array must be a two-dimensional iterable that
        specifies the components. As long as the components make sense
        in terms of compatibility of spaces, the input will be
        accepted.

        """
        from bempp_cl.api.assembly.boundary_operator import BoundaryOperator
        from collections.abc import Iterable

        def make_blocked(operator):
            """Turn a BoundaryOperator into a 1x1 blocked operator."""
            blocked_operator = BlockedOperator(1, 1)
            blocked_operator[0, 0] = operator
            return blocked_operator

        self._ops = []
        self._components_per_row = None
        self._components_per_column = None

        # First iterate through the array and transform each component into a
        # generalized blocked operator.

        for row in array:
            current_row = []
            for elem in row:
                if isinstance(elem, Iterable):
                    current_row.append(GeneralizedBlockedOperator(elem))
                elif isinstance(elem, BoundaryOperator):
                    current_row.append(make_blocked(elem))
                elif isinstance(elem, BlockedOperatorBase):
                    current_row.append(elem)
                else:
                    raise ValueError("Cannot process element of type: {0}".format(type(elem)))
            self._ops.append(current_row)

            all_domain_spaces = []
            all_range_spaces = []
            all_dual_to_range_spaces = []

            for row in self._ops:
                range_spaces = row[0].range_spaces
                dual_to_range_spaces = row[0].dual_to_range_spaces
                domain_spaces = []
                for elem in row:
                    if elem.range_spaces != range_spaces:
                        raise ValueError("Incompatible range spaces detected.")
                    if elem.dual_to_range_spaces != dual_to_range_spaces:
                        raise ValueError("Incompatible dual to range spaces detected.")
                    domain_spaces.extend(elem.domain_spaces)
                all_range_spaces.extend(range_spaces)
                all_dual_to_range_spaces.extend(dual_to_range_spaces)
                if all_domain_spaces:
                    # We have already processed one row
                    # and compare domain spaces to it.
                    if domain_spaces != all_domain_spaces:
                        raise ValueError("Incompatible domain spaces detected.")
                else:
                    # We are at the first row.
                    all_domain_spaces = domain_spaces

            self._domain_spaces = tuple(all_domain_spaces)
            self._dual_to_range_spaces = tuple(all_dual_to_range_spaces)
            self._range_spaces = tuple(all_range_spaces)

            super().__init__()

    def _assemble(self):
        """Implement the weak form."""
        assembled_list = []
        for row in self._ops:
            assembled_row = []
            for elem in row:
                assembled_row.append(elem.weak_form())
            assembled_list.append(assembled_row)
        return GeneralizedDiscreteBlockedOperator(assembled_list)

    @property
    def range_spaces(self):
        """Return the list of range spaces."""
        return tuple(self._range_spaces)

    @property
    def dual_to_range_spaces(self):
        """Return the list of dual_to_range spaces."""
        return tuple(self._dual_to_range_spaces)

    @property
    def domain_spaces(self):
        """Return the list of domain spaces."""
        return tuple(self._domain_spaces)


class MultitraceOperatorFromAssembler(BlockedOperatorBase):
    """A multitrace operator from an assembler."""

    def __init__(
        self,
        domain_spaces,
        range_spaces,
        dual_to_range_spaces,
        assembler,
        multitrace_operator_descriptor,
    ):
        """Generate operator from assembler."""
        self._domain_spaces = tuple(domain_spaces)
        self._range_spaces = tuple(range_spaces)
        self._dual_to_range_spaces = tuple(dual_to_range_spaces)
        self._assembler = assembler
        self._descriptor = multitrace_operator_descriptor

        super().__init__()

    def _assemble(self, *args, **kwargs):
        """Assemble the operator."""
        return self._assembler.assemble(self._descriptor, *args, **kwargs)

    @property
    def range_spaces(self):
        """Return a list of range spaces of the blocked operator."""
        return self._range_spaces

    @property
    def dual_to_range_spaces(self):
        """Return a list of dual to range spaces of the blocked operator."""
        return self._dual_to_range_spaces

    @property
    def domain_spaces(self):
        """Return a list of domain spaces of the blocked operator."""
        return self._domain_spaces


class SumBlockedOperator(BlockedOperatorBase):
    """Represents the sum of two blocked boundary operators."""

    def __init__(self, op1, op2):
        """Construct the object from two blocked operators."""
        if (
            op1.domain_spaces != op2.domain_spaces
            or op1.range_spaces != op2.range_spaces
            or op1.dual_to_range_spaces != op2.dual_to_range_spaces
        ):
            raise ValueError("Incompatible spaces for summation.")

        self._op1 = op1
        self._op2 = op2

        super().__init__()

    def _assemble(self):
        return self._op1.weak_form() + self._op2.weak_form()

    @property
    def range_spaces(self):
        """Return the list of range spaces."""
        return tuple(self._op1.range_spaces)

    @property
    def dual_to_range_spaces(self):
        """Return the list of dual_to_range spaces."""
        return tuple(self._op1.dual_to_range_spaces)

    @property
    def domain_spaces(self):
        """Return the list of domain spaces."""
        return tuple(self._op1.domain_spaces)


class ProductBlockedOperator(BlockedOperatorBase):
    """Represents the Product of two blocked boundary operators."""

    def __init__(self, op1, op2):
        """Construct the blocked operator product op1 * op2."""
        if op2.range_spaces != op1.domain_spaces:
            raise ValueError("Incompatible spaces for multiplication.")

        self._op1 = op1
        self._op2 = op2

        super().__init__()

    def _assemble(self):
        return self._op1.weak_form() * self._op2.strong_form()

    @property
    def range_spaces(self):
        """Return the list of range spaces."""
        return tuple(self._op1.range_spaces)

    @property
    def dual_to_range_spaces(self):
        """Return the list of dual_to_range spaces."""
        return tuple(self._op1.dual_to_range_spaces)

    @property
    def domain_spaces(self):
        """Return the list of domain spaces."""
        return tuple(self._op2.domain_spaces)


class ScaledBlockedOperator(BlockedOperatorBase):
    """Represents the scalar multiplication of a blocked operator."""

    def __init__(self, op, alpha):
        """Construct the scaled operator alpha * op."""
        self._op = op
        self._alpha = alpha

        super(ScaledBlockedOperator, self).__init__()

    def _assemble(self):
        """Assemble operator."""
        return self._alpha * self._op.weak_form()

    @property
    def range_spaces(self):
        """Return the list of range spaces."""
        return tuple(self._op.range_spaces)

    @property
    def dual_to_range_spaces(self):
        """Return the list of dual_to_range spaces."""
        return tuple(self._op.dual_to_range_spaces)

    @property
    def domain_spaces(self):
        """Return the list of domain spaces."""
        return tuple(self._op.domain_spaces)


class GeneralizedDiscreteBlockedOperator(_DiscreteOperatorBase):
    """A discrete generalized blocked operator."""

    def __init__(self, operators):
        """Initialize a generalized blocked operator."""
        from bempp_cl.api.utils.data_types import combined_type

        self._operators = operators

        shape = [0, 0]
        # Get column dimension
        for elem in operators[0]:
            shape[1] += elem.shape[1]
        # Get row dimension
        for row in operators:
            shape[0] += row[0].shape[0]

        shape = tuple(shape)

        # Get dtype

        dtype = operators[0][0].dtype
        for row in operators:
            for elem in row:
                dtype = combined_type(dtype, elem.dtype)

        # Sanity check of dimensions

        for row in operators:
            row_dim = row[0].shape[0]
            column_dim = 0
            for elem in row:
                if elem.shape[0] != row_dim:
                    raise ValueError("Incompatible dimensions detected.")
                column_dim += elem.shape[1]
            if column_dim != shape[1]:
                raise ValueError("Incompatible dimensions detected.")

        super().__init__(dtype, shape)

    def to_dense(self):
        """Return dense matrix."""
        rows = []
        for row in self._operators:
            rows.append([op.to_dense() for op in row])
        return _np.block(rows)

    def _matmat(self, other):
        """Implement the matrix/vector product."""
        from bempp_cl.api.utils.data_types import combined_type

        row_count = 0
        output = _np.zeros(
            (self.shape[0], other.shape[1]),
            dtype=combined_type(self.dtype, other.dtype),
        )

        for row in self._operators:
            row_dim = row[0].shape[0]
            column_count = 0
            for elem in row:
                output[row_count : row_count + row_dim, :] += (
                    elem @ other[column_count : column_count + elem.shape[1], :]
                )
                column_count += elem.shape[1]
            row_count += row_dim

        return output


class BlockedDiscreteOperator(_DiscreteOperatorBase):
    """Implementation of a discrete blocked boundary operator."""

    def __init__(self, ops):
        """
        Construct an operator from a two dimensional Numpy array of operators.

        ops is a list of list containing discrete boundary operators or None.
        A None entry is equivalent to a zero discrete boundary operator.

        """
        # pylint: disable=too-many-branches
        from bempp_cl.api.utils.data_types import combined_type
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            ZeroDiscreteBoundaryOperator,
        )

        if not isinstance(ops, _np.ndarray):
            ops = _np.array(ops)

        rows = ops.shape[0]
        cols = ops.shape[1]

        self._ndims = (rows, cols)

        self._operators = _np.empty((rows, cols), dtype=object)
        self._rows = -_np.ones(rows, dtype=int)
        self._cols = -_np.ones(cols, dtype=int)

        for i in range(rows):
            for j in range(cols):
                if ops[i, j] is None:
                    continue
                if self._rows[i] != -1:
                    if ops[i, j].shape[0] != self._rows[i]:
                        raise ValueError("Block row {0} has incompatible ".format(i) + " operator sizes.")
                else:
                    self._rows[i] = ops[i, j].shape[0]

                if self._cols[j] != -1:
                    if ops[i, j].shape[1] != self._cols[j]:
                        raise ValueError("Block column {0} has incompatible".format(j) + "operator sizes.")
                else:
                    self._cols[j] = ops[i, j].shape[1]
                self._operators[i, j] = ops[i, j]

        if not self._fill_complete():
            raise ValueError("Each row and column must contain at least one operator.")

        for i in range(rows):
            for j in range(cols):
                if self._operators[i, j] is None:
                    self._operators[i, j] = ZeroDiscreteBoundaryOperator(self._rows[i], self._cols[j])

        shape = (_np.sum(self._rows), _np.sum(self._cols))

        dtype = "float32"
        for obj in self._operators.ravel():
            if obj is not None:
                dtype = combined_type(dtype, obj.dtype)

        super().__init__(dtype, shape)

    def __getitem__(self, key):
        """Return the object at position (i, j)."""
        return self._operators[key]

    def _fill_complete(self):
        if (-1 in self._rows) or (-1 in self._cols):
            return False
        return True

    def _matvec(self, x):
        from bempp_cl.api.utils.data_types import combined_type

        if not self._fill_complete():
            raise ValueError("Not all rows or columns contain operators.")

        if x.ndim > 1:
            return self._matmat(x)

        row_dim = 0
        res = _np.zeros(self.shape[0], dtype=combined_type(self.dtype, x.dtype))

        for i in range(self._ndims[0]):
            col_dim = 0
            local_res = res[row_dim : row_dim + self._rows[i]]
            for j in range(self._ndims[1]):
                local_x = x[col_dim : col_dim + self._cols[j]]
                op_is_complex = _np.iscomplexobj(self._operators[i, j].dtype.type(1))
                if _np.iscomplexobj(x) and not op_is_complex:
                    local_res += self._operators[i, j].dot(_np.real(local_x)) + 1j * self._operators[i, j].dot(
                        _np.imag(local_x)
                    )
                else:
                    local_res += self._operators[i, j].dot(local_x)
                col_dim += self._cols[j]
            row_dim += self._rows[i]

        return res

    def _matmat(self, x):
        from bempp_cl.api.utils.data_types import combined_type

        if not self._fill_complete():
            raise ValueError("Not all rows or columns contain operators.")

        row_dim = 0
        res = _np.zeros((self.shape[0], x.shape[1]), dtype=combined_type(self.dtype, x.dtype))

        for i in range(self._ndims[0]):
            col_dim = 0
            local_res = res[row_dim : row_dim + self._rows[i], :]
            for j in range(self._ndims[1]):
                local_x = x[col_dim : col_dim + self._cols[j], :]
                op_is_complex = _np.iscomplexobj(self._operators[i, j].dtype.type(1))
                if _np.iscomplexobj(x) and not op_is_complex:
                    local_res[:] += self._operators[i, j].dot(_np.real(local_x)) + 1j * self._operators[i, j].dot(
                        _np.imag(local_x)
                    )
                else:
                    local_res[:] += self._operators[i, j].dot(local_x)
                col_dim += self._cols[j]
            row_dim += self._rows[i]
        return res

    def _get_row_dimensions(self):
        return self._rows

    def _get_column_dimensions(self):
        return self._cols

    def to_dense(self):
        """Return dense matrix."""
        rows = []
        for i in range(self._ndims[0]):
            row = [self[i, j].to_dense() for j in range(self._ndims[1])]
            rows.append(_np.hstack(row))
        return _np.vstack(rows)

    def _transpose(self):
        """Implement the transpose."""
        raise NotImplementedError()

    def _adjoint(self):
        """Implement the adjoint."""
        raise NotImplementedError()

    row_dimensions = property(_get_row_dimensions)
    column_dimensions = property(_get_column_dimensions)


# pylint: disable=invalid-name
def coefficients_from_grid_functions_list(grid_funs):
    """
    Return a vector of coefficients of a list of grid functions.

    Given a list [f0, f1, f2, ...] this function returns a
    single Numpy vector containing [f1.coefficients, f2.coefficients, ...].

    Parameters
    ----------
    grid_funs : list of GridFunction objects
        A list containing the grid functions
    """
    vec_len = 0
    input_type = _np.dtype("float32")
    for item in grid_funs:
        input_type = _np.promote_types(input_type, item.coefficients.dtype)
        vec_len += item.space.global_dof_count
    res = _np.zeros(vec_len, dtype=input_type)
    pos = 0
    for item in grid_funs:
        dof_count = item.space.global_dof_count
        res[pos : pos + dof_count] = item.coefficients
        pos += dof_count
    return res


def projections_from_grid_functions_list(grid_funs, projection_spaces):
    """
    Return a vector of projections of a list of grid functions.

    Given a list [f0, f1, f2, ...] this function returns a
    single Numpy vector containing
    [f0.projections(projection_spaces[0]),
     f1.projections(projection_spaces[1])],
     ...].

    Parameters
    ----------
    grid_funs : list of GridFunction objects
        A list containing the grid functions
    projection_spaces : list of projection spaces
        A list of projection spaces. Must have the same
        length as grid_funs.
    """
    projections = []
    for item, proj_space in zip(grid_funs, projection_spaces):
        projections.append(item.projections(proj_space))
    vec_len = 0
    input_type = _np.dtype("float32")
    for item in projections:
        input_type = _np.promote_types(input_type, item.dtype)
        vec_len += len(item)
    res = _np.zeros(vec_len, dtype=input_type)
    pos = 0
    for item in projections:
        dof_count = len(item)
        res[pos : pos + dof_count] = item
        pos += dof_count
    return res


# pylint: disable=invalid-name
def grid_function_list_from_coefficients(coefficients, spaces):
    """
    Create a list of grid functions from a long vector of coefficients.

    Parameters
    ----------
    coefficients : np.ndarray
        One-dimensional array of coefficients
    spaces : list of Space objects
        The sum of the global dofs of the spaces must be equal to the
        length of the coefficients vector.
    """
    from bempp_cl.api import GridFunction

    pos = 0
    res_list = []
    for space in spaces:
        dof_count = space.global_dof_count
        res_list.append(GridFunction(space, coefficients=coefficients[pos : pos + dof_count]))
        pos += dof_count
    return res_list


# pylint: disable=invalid-name
def grid_function_list_from_projections(projections, spaces, dual_spaces=None):
    """
    Create a list of grid functions from a long vector of projections.

    Parameters
    ----------
    coefficients : np.ndarray
        One-dimensional array of coefficients
    spaces : list of Space objects
        The sum of the global dofs of the spaces must be equal to the
        length of the coefficients vector.
    dual_spaces : list of Space objects
        The associated dual spaces. If None use the spaces as dual spaces.
    """
    from bempp_cl.api import GridFunction

    pos = 0
    res_list = []
    if dual_spaces is None:
        dual_spaces = spaces
    if len(spaces) != len(dual_spaces):
        raise ValueError("spaces must have the same length as dual_spaces")
    for space, dual in zip(spaces, dual_spaces):
        dof_count = space.global_dof_count
        res_list.append(GridFunction(space, projections=projections[pos : pos + dof_count], dual_space=dual))
        pos += dof_count
    return res_list



"""Maxwell far-field operators."""

import numpy as _np


def electric_field(
    space,
    points,
    wavenumber,
    parameters=None,
    assembler="dense",
    device_interface=None,
    precision=None,
):
    """Return a Maxwell electric far-field potential operator."""
    import bempp_cl.api
    from bempp_cl.api.operators import OperatorDescriptor
    from bempp_cl.api.assembly.potential_operator import PotentialOperator
    from bempp_cl.api.assembly.assembler import PotentialAssembler

    if precision is None:
        precision = bempp_cl.api.DEFAULT_PRECISION

    operator_descriptor = OperatorDescriptor(
        "maxwell_far_field_electric_field_potential",  # Identifier
        [_np.real(wavenumber), _np.imag(wavenumber)],  # Options
        "helmholtz_far_field_single_layer",  # Kernel type
        "maxwell_electric_far_field",  # Assembly type
        precision,  # Precision
        True,  # Is complex
        None,  # Singular part
        3,  # Kernel dimension
    )

    return PotentialOperator(
        PotentialAssembler(space, points, operator_descriptor, device_interface, assembler, parameters)
    )


def magnetic_field(
    space,
    points,
    wavenumber,
    parameters=None,
    assembler="dense",
    device_interface=None,
    precision=None,
):
    """Return a Maxwell magnetic far-field potential operator."""
    import bempp_cl.api
    from bempp_cl.api.operators import OperatorDescriptor
    from bempp_cl.api.assembly.potential_operator import PotentialOperator
    from bempp_cl.api.assembly.assembler import PotentialAssembler

    if precision is None:
        precision = bempp_cl.api.DEFAULT_PRECISION

    operator_descriptor = OperatorDescriptor(
        "maxwell_far_field_magnetic_field_potential",  # Identifier
        [_np.real(wavenumber), _np.imag(wavenumber)],  # Options
        "helmholtz_far_field_single_layer",  # Kernel type
        "maxwell_magnetic_far_field",  # Assembly type
        precision,  # Precision
        True,  # Is complex
        None,  # Singular part
        3,  # Kernel dimension
    )

    return PotentialOperator(
        PotentialAssembler(space, points, operator_descriptor, device_interface, assembler, parameters)
    )


# def magnetic_field(
# space, points, wavenumber, parameters=None, device_interface=None, precision=None
# ):
# """Return a Maxwell magnetic field potential operator."""
# from bempp_cl.core.dense_potential_assembler import DensePotentialAssembler
# from bempp_cl.api.operators import OperatorDescriptor
# from bempp_cl.api.operators import _add_wavenumber
# from bempp_cl.api.assembly.potential_operator import PotentialOperator

# options = {}
# options["KERNEL_FUNCTION"] = "helmholtz_gradient"
# _add_wavenumber(options, wavenumber)

# return PotentialOperator(
# DensePotentialAssembler(
# space,
# OperatorDescriptor(
# "maxwell_magnetic_field_potential", options, "maxwell_magnetic_field"
# ),
# points,
# 3,
# True,
# device_interface,
# precision,
# parameters=parameters,
# )
# )



"""Various assemblers to discretize boundary operators."""


def _create_assembler(domain, dual_to_range, identifier, parameters, device_interface=None):
    """Create assembler based on string."""
    from bempp_cl.core.singular_assembler import SingularAssembler
    from bempp_cl.core.dense_assembler import DenseAssembler
    from bempp_cl.core.diagonal_assembler import DiagonalAssembler
    from bempp_cl.api.fmm.fmm_assembler import FmmAssembler
    from bempp_cl.api import check_for_fmm

    # from bempp_cl.core.numba.dense_assembler import DenseAssembler
    from bempp_cl.core.sparse_assembler import SparseAssembler

    # from bempp_cl.core.dense_assembler import DenseAssembler
    # from bempp_cl.core.sparse_assembler import SparseAssembler
    # from bempp_cl.core.dense_evaluator import DenseEvaluatorAssembler
    # from bempp_cl.core.dense_multitrace_evaluator import DenseMultitraceEvaluatorAssembler

    if identifier == "only_singular_part":
        return SingularAssembler(domain, dual_to_range, parameters)
    if identifier == "only_diagonal_part":
        return DiagonalAssembler(domain, dual_to_range, parameters)
    if identifier == "dense":
        return DenseAssembler(domain, dual_to_range, parameters)
    if identifier == "default_nonlocal":
        return DenseAssembler(domain, dual_to_range, parameters)
    if identifier == "sparse":
        return SparseAssembler(domain, dual_to_range, parameters)
    if identifier == "fmm":
        if not check_for_fmm():
            raise ValueError("No compatible FMM library found. Please install Exafmm from github.com/exafmm/exafmm-t.")
        return FmmAssembler(domain, dual_to_range, parameters)
    else:
        raise ValueError("Unknown assembler type.")
    # if identifier == "dense_evaluator":
    # return DenseEvaluatorAssembler(domain, dual_to_range, parameters)
    # if identifier == "multitrace_evaluator":
    # return DenseMultitraceEvaluatorAssembler(domain, dual_to_range, parameters)


class AssemblerInterface(object):
    """Default Assembler interface object."""

    def __init__(
        self,
        domain,
        dual_to_range,
        assembler,
        device_interface,
        precision,
        parameters=None,
    ):
        """Initialize assembler based on assembler_type string."""
        import bempp_cl.api
        import bempp_cl.api as _api

        self._domain = domain
        self._dual_to_range = dual_to_range
        self._parameters = _api.assign_parameters(parameters)
        self._device_interface = device_interface
        self._precision = precision

        if self._device_interface is None:
            self._device_interface = bempp_cl.api.DEFAULT_DEVICE_INTERFACE

        if not isinstance(assembler, str):
            self._implementation = assembler
        else:
            self._implementation = _create_assembler(
                domain, dual_to_range, assembler, self.parameters, device_interface
            )

    @property
    def domain(self):
        """Return domain space."""
        return self._domain

    @property
    def dual_to_range(self):
        """Return dual to range space."""
        return self._dual_to_range

    @property
    def parameters(self):
        """Return parameters."""
        return self._parameters

    def assemble(self, operator_descriptor, *args, **kwargs):
        """Assemble the operator."""
        return self._implementation.assemble(
            operator_descriptor,
            self._device_interface,
            self._precision,
            *args,
            **kwargs,
        )


class AssemblerBase(object):
    """Base class for assemblers."""

    def __init__(self, domain, dual_to_range, parameters=None):
        """Instantiate the base class."""
        import bempp_cl.api as api

        self._domain = domain
        self._dual_to_range = dual_to_range
        self._parameters = api.assign_parameters(parameters)

    @property
    def domain(self):
        """Return domain."""
        return self._domain

    @property
    def dual_to_range(self):
        """Return dual to range."""
        return self._dual_to_range

    @property
    def parameters(self):
        """Return parameters."""
        return self._parameters

    def assemble(self, operator_descriptor, *args, **kwargs):
        """Assemble the operator."""
        raise NotImplementedError("Needs to be implemented by derived class.")


class PotentialAssembler(object):
    """Base class for potential assemblers."""

    def __init__(
        self,
        space,
        points,
        operator_descriptor,
        device_interface,
        assembler,
        parameters,
    ):
        """Interface for potential operators."""
        self.space = space
        self.points = points
        self.kernel_dimension = operator_descriptor.kernel_dimension
        self._is_complex = operator_descriptor.is_complex

        self._implementation = select_potential_implementation(
            space, points, operator_descriptor, device_interface, assembler, parameters
        )

    def evaluate(self, x):
        """Evaluate the potential."""
        import numpy as np

        if not self._is_complex:
            if np.iscomplexobj(x):
                return self._implementation.evaluate(np.real(x)) + 1j * self._implementation.evaluate(np.imag(x))
            else:
                return self._implementation.evaluate(x)
        else:
            return self._implementation.evaluate(x)


def select_potential_implementation(space, points, operator_descriptor, device_interface, assembler, parameters):
    """Select a potential operator implementation."""
    import bempp_cl.api

    parameters = bempp_cl.api.assign_parameters(parameters)
    if device_interface is None:
        device_interface = bempp_cl.api.DEFAULT_DEVICE_INTERFACE

    if assembler == "dense":
        from bempp_cl.core.dense_potential_assembler import DensePotentialAssembler

        return DensePotentialAssembler(space, operator_descriptor, points, device_interface, parameters)
    elif assembler == "fmm":
        from bempp_cl.api.fmm.fmm_assembler import FmmPotentialAssembler

        if not bempp_cl.api.check_for_fmm():
            raise ValueError("No compatible FMM library found. Please install Exafmm from github.com/exafmm/exafmm-t.")

        return FmmPotentialAssembler(space, operator_descriptor, points, device_interface, parameters)
    else:
        raise ValueError(f"Unknown potential assembler: {assembler}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep  3 10:40:49 2024

@author: tom
"""

from numpy import array, zeros, ones, pi, sqrt, tan, sin
from electroacPy.general.freqop import laplace


class pressureSource:
    def __init__(self, np, nm, value):
        """
        Create a resistor component.

        Parameters
        ----------
        np : int
            positive node.
        nm : int
            negative node.
        value : float
            resustance value.

        Returns
        -------
        None.

        """
        
        self.np = str(np)
        self.nm = str(nm)
        self.value = value
        self.G = 1
        self.Gs = None
        
        # create stamp and relative informations
        self.stamp_G = array([[0, 0, 1], [0, 0, -1], [1, -1, 0]])
        self.stamp_I = array([[0], [0], [value]])
        self.contribute = ["G", "Is"]  
        self.vsource = 1
    
        
    def init_component(self, frequency):
        self.Gs = 1 * ones(len(frequency)) 
        
    
    def update_stamp(self, node_id, M, nbsource):
        """
        Update the component's stamp in a matrix of the system's size.

        Parameters
        ----------
        node_list : list
            list of nodes.

        Returns
        -------
        None.

        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        maxNode = max(node_id.values())
        
        # G stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np != 0:
            self.stamp_G[maxNode+nbsource, np-1] = 1  # sub mat B
            self.stamp_G[np-1, maxNode+nbsource] = 1  # sub mat C
        if nm != 0:
            self.stamp_G[maxNode+nbsource, nm-1] = -1 # sub mat B
            self.stamp_G[nm-1, maxNode+nbsource] = -1 # sub mat C     
        
        # I stamp
        self.stamp_I = zeros([maxNode+M, 1], dtype=complex) 
        self.stamp_I[maxNode+nbsource] = self.value #1


#%% loudspeaker relared components
class radiator:
    def __init__(self, np, nm, Sd, rho=1.22, c=343):
        """
        Create a resistor component.

        Parameters
        ----------
        np : int
            positive node.
        nm : int
            negative node.
        value : float
            resustance value.

        Returns
        -------
        None.

        """
        
        self.np = str(np)
        self.nm = str(nm)
        self.Sd = Sd
        self.G = 1/Sd
        self.Gs = None
        
        self.rho = rho
        self.c = c
        self.r = sqrt(Sd/pi)
        
        # create stamp and relative informations
        self.stamp_G = array([[1, -1], [-1, 1]])
        self.contribute = ["G"]        
        self.vsource = 0

    def init_component(self, frequency):
        """
        initialize resistor within circuit

        Parameters
        ----------
        frequency : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        # self.Gs = self.G * ones(len(frequency)) 
        from scipy.special import j1, struve
        
        om = 2 * pi * frequency
        k = om / self.c
        Zc = self.rho * self.c

        self.Za = (Zc*(1 - j1(2 * k * self.r) / k / self.r) +
                    1j * (Zc*struve(1, 2 * k * self.r) / k / self.r))
        self.Gs = 1/self.Za # conductance
        
        
    def update_stamp(self, node_id, M):
        """
        Update the component's stamp in a matrix of the system's size.
 
        Parameters
        ----------
        node_list : list
            list of nodes.
 
        Returns
        -------
        None.
 
        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        maxNode = max(node_id.values())
        
        # update stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np != 0:
            self.stamp_G[np - 1, np - 1] = 1
        if nm != 0:
            self.stamp_G[nm - 1, nm - 1] = 1
        if np != 0 and nm != 0:
            self.stamp_G[np - 1, nm - 1] = -1
            self.stamp_G[nm - 1, np - 1] = -1
    
class cavity:
    def __init__(self, np, nm, Vb, eta=1e-5, rho=1.22, c=343, losses=True):
        """
        Create a resistor component.

        Parameters
        ----------
        np : int
            positive node.
        nm : int
            negative node.
        value : float
            resustance value.

        Returns
        -------
        None.

        """
        
        self.np = str(np)
        self.nm = str(nm)
        self.Vb = Vb
        self.eta = eta
        self.rho = rho
        self.c = c
        self.losses = losses
        self.G = 1
        self.Gs = None
        
        # create stamp and relative informations
        self.stamp_G = array([[1, -1], [-1, 1]])
        self.contribute = ["G"]        
        self.vsource = 0

    def init_component(self, frequency):
        """
        initialize resistor within circuit

        Parameters
        ----------
        frequency : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        s = laplace(frequency)
        # Lb = (self.Vb)**(1/3) / 3
        # Sb = self.Vb / Lb
        
        Cb = self.Vb / self.rho / self.c**2
        Rb = self.rho * self.c / self.eta / self.Vb
        if self.losses is True:
            self.Gs = s * Cb + 1/Rb # conductance
        elif self.losses is False:
            self.Gs = s * Cb
        else:
            raise Exception("Losses must be boolean.")
        
        
    def update_stamp(self, node_id, M):
        """
        Update the component's stamp in a matrix of the system's size.
 
        Parameters
        ----------
        node_list : list
            list of nodes.
 
        Returns
        -------
        None.
 
        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        maxNode = max(node_id.values())
        
        # update stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np != 0:
            self.stamp_G[np - 1, np - 1] = 1
        if nm != 0:
            self.stamp_G[nm - 1, nm - 1] = 1
        if np != 0 and nm != 0:
            self.stamp_G[np - 1, nm - 1] = -1
            self.stamp_G[nm - 1, np - 1] = -1

class port:
    def __init__(self, np, nm, Lp, rp, flange="single", 
                mu=1.86e-5, rho=1.22, c=343):
        """
        Create a resistor component.

        Parameters
        ----------
        np : int
            positive node.
        nm : int
            negative node.
        value : float
            resustance value.

        Returns
        -------
        None.

        """
        
        self.np = str(np)
        self.nm = str(nm)
        self.Lp = Lp
        self.rp = rp
        self.Sp = pi * rp**2
        self.rho = rho
        self.c = c
        self.mu = mu
        self.G = 1
        self.Gs = None
        
        # prepare flange
        if flange == "single":
            self.flangeCoeff = 0.84
        elif flange == "both":
            self.flangeCoeff = 0.96
        elif flange == "none":
            self.flangeCoeff = 0.76
        else:
            self.flangeCoeff = 0.84
        
        # create stamp and relative informations
        self.stamp_G = array([[1, -1], [-1, 1]])
        self.contribute = ["G"]        
        self.vsource = 0

    def init_component(self, frequency):
        """
        initialize resistor within circuit

        Parameters
        ----------
        frequency : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        om = 2 * pi * frequency
        s = laplace(frequency)
        
        Lt = self.Lp + self.flangeCoeff * sqrt(self.Sp)
        
        Mp = Lt*self.rho / (self.Sp)
        Rp = (2*om*self.rho*self.mu)/self.Sp * (Lt/self.rp + 1*0.7)
        Zp = s*Mp + Rp
        self.Gs = 1 / Zp # conductance
        
        
    def update_stamp(self, node_id, M):
        """
        Update the component's stamp in a matrix of the system's size.
 
        Parameters
        ----------
        node_list : list
            list of nodes.
 
        Returns
        -------
        None.
 
        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        maxNode = max(node_id.values())
        
        # update stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np != 0:
            self.stamp_G[np - 1, np - 1] = 1
        if nm != 0:
            self.stamp_G[nm - 1, nm - 1] = 1
        if np != 0 and nm != 0:
            self.stamp_G[np - 1, nm - 1] = -1
            self.stamp_G[nm - 1, np - 1] = -1    

class membrane:
    def __init__(self, np, nm, Cm, Mm, Rm, Sd, rho=1.22, c=343):
        self.np = str(np)
        self.nm = str(nm)
        
        self.Cm = Cm
        self.Mm = Mm
        self.Rm = Rm
        self.Sd = Sd
        self.rho = rho
        self.c = c
        self.G = 1
        self.Gs = None
        
        self.Ma = Mm / Sd**2
        self.Ca = Cm * Sd**2
        self.Ra = Rm / Sd**2
        
        # create stamp and relative informations
        self.stamp_G = array([[1, -1], [-1, 1]])
        self.contribute = ["G"]        
        self.vsource = 0
   
    def init_component(self, frequency):
       """
       initialize resistor within circuit

       Parameters
       ----------
       frequency : TYPE
           DESCRIPTION.

       Returns
       -------
       None.

       """
       s = laplace(frequency)
       
       Za = s*self.Ma + 1/self.Ca/s + self.Ra
       self.Gs = 1 / Za # conductance
       
       
    def update_stamp(self, node_id, M):
        """
        Update the component's stamp in a matrix of the system's size.
 
        Parameters
        ----------
        node_list : list
            list of nodes.
 
        Returns
        -------
        None.
 
        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        maxNode = max(node_id.values())
        
        # update stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np != 0:
            self.stamp_G[np - 1, np - 1] = 1
        if nm != 0:
            self.stamp_G[nm - 1, nm - 1] = 1
        if np != 0 and nm != 0:
            self.stamp_G[np - 1, nm - 1] = -1
            self.stamp_G[nm - 1, np - 1] = -1   
            
class closed_line:
    def __init__(self, np, nm, Lp, Sp, eta=1e-3, rho=1.22, c=343):
        """
        Add a closed cavity (nm linked to ground)

        Parameters
        ----------
        np : TYPE
            DESCRIPTION.
        nm : TYPE
            DESCRIPTION.
        Lp : TYPE
            DESCRIPTION.
        Sp : TYPE
            DESCRIPTION.
        rho : TYPE, optional
            DESCRIPTION. The default is 1.22.
        c : TYPE, optional
            DESCRIPTION. The default is 343.

        Returns
        -------
        None.

        """
        
        self.np = str(np)
        self.nm = str(nm)
        
        self.Lp = Lp
        self.Sp = Sp
        self.rp = sqrt(Sp / pi) # equivalent radius (not necessarily circular duct)
        self.Pp = 2*pi*self.rp  # equivalent perimeter (to estimate thermo-viscous losses)
        self.rho = rho
        self.c = c
        self.G = 1
        self.Gs = None
        self.eta = eta
        self.stamp_G = None
        self.contribute = ["G"]        
        
        self.vsource = 0
        
    def init_component(self, frequency):
       """
       initialize resistor within circuit

       Parameters
       ----------
       frequency : TYPE
           DESCRIPTION.

       Returns
       -------
       None.

       """
       om = 2 * pi * frequency
       k = om / self.c
       Zc = self.rho * self.c / self.Sp
       kl = k-1j*self.eta # kloss(self.Sp, self.Pp, k, self.rho, self.c)
       Za = Zc / 1j / tan(kl*self.Lp)
       self.Gs = 1 / Za
       
       
       
    def update_stamp(self, node_id, M):
        """
        Update the component's stamp in a matrix of the system's size.
 
        Parameters
        ----------
        node_list : list
            list of nodes.
 
        Returns
        -------
        None.
 
        """
        if self.np != '0':
            np = node_id[self.np]
        else:
            np = 0
            
        if self.nm != '0':
            nm = node_id[self.nm]
        else:
            nm = 0
        
        maxNode = max(node_id.values())
        
        # update stamp
        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np != 0:
            self.stamp_G[np - 1, np - 1] = 1
        if nm != 0:
            self.stamp_G[nm - 1, nm - 1] = 1
        if np != 0 and nm != 0:
            self.stamp_G[np - 1, nm - 1] = -1
            self.stamp_G[nm - 1, np - 1] = -1


class open_line_T:
    def __init__(self, np0, np1, np2, nm, Lp, Sp, eta=1e-3, rho=1.22, c=343):
        self.np = str(np0) # careful with numbering and references
        self.nm = str(np1)
        self.np1 = str(np2)
        self.nm1 = str(nm)  # ground
        self.Lp = Lp
        self.Sp = Sp  # surface
        self.rp = sqrt(Sp / pi)  # equivalent radius (not necessarily circular duct)
        self.Pp = 2 * pi * self.rp  # equivalent perimeter (to estimate thermo-viscous losses)
        self.rho = rho
        self.c = c   
        self.eta = eta
        
        self.G = 1
        self.Gs = None
        self.Ys = None
        self.stamp_G = None
        self.stamp_Y = None
        self.contribute = ["G"]        
        
        self.vsource = 0


    def init_component(self, frequency):
       """
       initialize resistor within circuit

       Parameters
       ----------
       frequency : TYPE
           DESCRIPTION.

       Returns
       -------
       None.

       """
       
       LL = self.Lp #+ 0.73*self.rp*2
       om = 2 * pi * frequency
       k = om / self.c
       Zc = self.rho * self.c / self.Sp
       kl = k - 1j*self.eta #kloss(self.Sp, self.Pp, k, self.rho, self.c) # k - 1j*self.eta #
       Zt =  1j * Zc * tan(kl*LL/2)
       Yt = 1j * sin(kl*LL) / Zc
       self.Gs = 1 / Zt
       self.Ys = Yt

    def update_stamp(self, node_id, M):
        """
        Update the component's stamp in a matrix of the system's size.
 
        Parameters
        ----------
        node_list : list
            list of nodes.
 
        Returns
        -------
        None.
 
        """
      
        
        if self.np != '0':
            np0 = node_id[self.np]
        else:
            np0 = 0
            
        if self.nm != '0':
            np1 = node_id[self.nm]
        else:
            np1 = 0
        
        if self.np1 != '0':
            np2 = node_id[self.np1]
        else:
            np2 = 0
            
        if self.nm1 != '0':
            nm = node_id[self.nm1]
        else:
            nm = 0
        
        maxNode = max(node_id.values())                

        self.stamp_G = zeros([maxNode+M, maxNode+M], dtype=complex)
        self.stamp_Y = zeros([maxNode+M, maxNode+M], dtype=complex)
        
        if np0 != 0:
            self.stamp_G[np0 - 1, np0 - 1] = 1
        if np1 != 0:
            self.stamp_G[np1 - 1, np1 - 1] = 2
            self.stamp_Y[np1 - 1, np1 - 1] = 1
        if np2 != 0:
            self.stamp_G[np2 - 1, np2 - 1] = 1
        if nm != 0:
            self.stamp_Y[nm - 1, nm - 1] = 1
            
        if np0 != 0 and np1 != 0:
            self.stamp_G[np0 - 1, np1 - 1] = -1
            self.stamp_G[np1 - 1, np0 - 1] = -1 
        
        if np1 != 0 and np2 != 0:
            self.stamp_G[np1 - 1, np2 - 1] = -1
            self.stamp_G[np2 - 1, np1 - 1] = -1 
            
        if np1 != 0 and nm !=0:
            self.stamp_Y[np1 - 1, nm - 1] = -1
            self.stamp_Y[nm - 1, np1 - 1] = -1 


def kloss(section, perimeter, k, rho=1.22, c=343):
    """
    Losses as expressed by Jean-Pierre Dalmont in his university courses

    Input parameters
    -----
    section: duct cross-sectional area
    perimeter: perimeter of the cross-section
    k: input wavenumber

    ------
    return: k_l --> wavenumber with losses
    """
    f = k * c / 2 / pi
    om = 2 * pi * f
    alpha = sqrt(f) * (0.95e-5 + 2.03e-5) * perimeter / 2 / section  # +1 was added
    k_l = om / c * (1 + (1 - 1j) * alpha)
    return k_l





"""Implementation of an FMM Assembler."""

from bempp_cl.api.assembly import assembler as _assembler
import numba as _numba
import numpy as _np

_FMM_CACHE = {}
_FMM_POTENTIAL_CACHE = {}


def get_mode_from_operator_identifier(identifier):
    """Get the Fmm mode from the operator identifier."""
    descriptor = identifier.split("_")[0]

    if descriptor == "laplace":
        return "laplace"
    elif descriptor == "helmholtz":
        return "helmholtz"
    elif descriptor == "modified":
        return "modified_helmholtz"
    elif descriptor == "maxwell":
        return "helmholtz"
    else:
        raise ValueError("Unknown identifier string.")


def get_fmm_interface(domain, dual_to_range, mode, wavenumber, parameters=None, device_interface=None):
    """Get an Fmm instance."""
    import bempp_cl.api

    global _FMM_CACHE

    parameters = bempp_cl.api.assign_parameters(parameters)

    key = (
        domain.grid.id,
        dual_to_range.grid.id,
        mode,
        wavenumber,
        parameters.fmm.expansion_order,
        parameters.fmm.ncrit,
    )

    interface = _FMM_CACHE.get(key, None)

    if interface is None:
        from bempp_cl.api.fmm.exafmm import ExafmmInterface

        interface = ExafmmInterface.from_grid(
            domain.grid,
            mode,
            wavenumber=wavenumber,
            target_grid=dual_to_range.grid,
            parameters=parameters,
            device_interface=device_interface,
        )
        _FMM_CACHE[key] = interface
    else:
        bempp_cl.api.log("Using cached Fmm Interface.", level="debug")

    return interface


def get_fmm_potential_interface(space, points, mode, wavenumber):
    """Get an Fmm potential instance."""
    import bempp_cl.api

    global _FMM_POTENTIAL_CACHE

    points_hash = hash(points.data.tobytes())

    key = (space.grid.id, points_hash, mode, wavenumber)

    interface = _FMM_POTENTIAL_CACHE.get(key, None)

    if interface is None:
        from bempp_cl.api.fmm.exafmm import ExafmmInterface

        quadrature_order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular

        interface = ExafmmInterface(
            space.grid.map_to_point_cloud(quadrature_order, precision="double"),
            points.T,
            mode,
            wavenumber,
            bempp_cl.api.GLOBAL_PARAMETERS.fmm.depth,
            bempp_cl.api.GLOBAL_PARAMETERS.fmm.expansion_order,
            bempp_cl.api.GLOBAL_PARAMETERS.fmm.ncrit,
        )
        _FMM_POTENTIAL_CACHE[key] = interface
    else:
        bempp_cl.api.log("Using cached Fmm Interface.", level="debug")

    return interface


def create_evaluator(operator_descriptor, fmm_interface, domain, dual_to_range, parameters):
    """Return an Fmm evaluator for the requested kernel."""
    if operator_descriptor.assembly_type == "default_scalar":
        return make_default_scalar(operator_descriptor, fmm_interface, domain, dual_to_range)
    if operator_descriptor.assembly_type.split("_")[-1] == "hypersingular":
        return make_scalar_hypersingular(operator_descriptor, fmm_interface, domain, dual_to_range)
    if operator_descriptor.assembly_type == "maxwell_electric_field":
        return make_maxwell_electric_field_boundary(operator_descriptor, fmm_interface, domain, dual_to_range)
    if operator_descriptor.assembly_type == "maxwell_magnetic_field":
        return make_maxwell_magnetic_field_boundary(operator_descriptor, fmm_interface, domain, dual_to_range)


def create_potential_evaluator(operator_descriptor, fmm_interface, space, parameters):
    """Select an Fmm Potential Evaluator."""
    if operator_descriptor.assembly_type == "default_scalar":
        return make_default_scalar_potential(operator_descriptor, fmm_interface, space)
    elif operator_descriptor.assembly_type == "maxwell_electric_field":
        return make_maxwell_electric_field_potential(operator_descriptor, fmm_interface, space)
    elif operator_descriptor.assembly_type == "maxwell_magnetic_field":
        return make_maxwell_magnetic_field_potential(operator_descriptor, fmm_interface, space)
    else:
        raise ValueError("Unknown descriptor.")


class FmmPotentialAssembler(object):
    """Potential assembler for FMM."""

    def __init__(self, space, operator_descriptor, points, device_interface, parameters):
        """Initialise FMM Potential Assembler."""
        mode = get_mode_from_operator_identifier(operator_descriptor.identifier)

        if mode == "laplace":
            wavenumber = None
        elif mode == "helmholtz":
            wavenumber = operator_descriptor.options[0] + 1j * operator_descriptor.options[1]
        elif mode == "modified_helmholtz":
            wavenumber = operator_descriptor.options[0]
        else:
            raise ValueError(f"Unknown value {mode} for `mode`.")

        fmm_potential_interface = get_fmm_potential_interface(space, points, mode, wavenumber)
        self._evaluator = create_potential_evaluator(operator_descriptor, fmm_potential_interface, space, parameters)

    def evaluate(self, x):
        """Actually evaluate the potential."""
        return self._evaluator(x)


class FmmAssembler(_assembler.AssemblerBase):
    """Assembler for Fmm."""

    # pylint: disable=useless-super-delegation
    def __init__(self, domain, dual_to_range, parameters):
        """Create an Fmm assembler instance."""
        super().__init__(domain, dual_to_range, parameters)

        # self._source_map = domain.map_to_points(parameters.quadrature.regular)
        # self._target_map = dual_to_range.map_to_points(
        # parameters.quadrature.regular, return_transpose=True
        # )

        self.dtype = None
        self._evaluator = None
        self.shape = (dual_to_range.global_dof_count, domain.global_dof_count)

    def assemble(self, operator_descriptor, device_interface, precision, *args, **kwargs):
        """Actually assemble."""
        from bempp_cl.api.space.space import return_compatible_representation
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            GenericDiscreteBoundaryOperator,
        )

        actual_domain, actual_dual_to_range = return_compatible_representation(self.domain, self.dual_to_range)

        mode = get_mode_from_operator_identifier(operator_descriptor.identifier)

        if mode == "laplace":
            wavenumber = None
        elif mode == "helmholtz":
            wavenumber = operator_descriptor.options[0] + 1j * operator_descriptor.options[1]
        elif mode == "modified_helmholtz":
            wavenumber = operator_descriptor.options[0]
        else:
            raise ValueError(f"Unknown value {mode} for `mode`.")

        fmm_interface = get_fmm_interface(
            actual_domain,
            actual_dual_to_range,
            mode,
            wavenumber,
            self.parameters,
            device_interface,
        )

        self._evaluator = create_evaluator(
            operator_descriptor,
            fmm_interface,
            actual_domain,
            actual_dual_to_range,
            self.parameters,
        )

        if operator_descriptor.is_complex:
            self.dtype = "complex128"
        else:
            self.dtype = "float64"

        return GenericDiscreteBoundaryOperator(self)

    def matvec(self, x):
        """Perform a matvec."""
        ndim = len(x.shape)

        if ndim > 2:
            raise ValueError("x must have shape (N, ) or (N, 1), where N is number of elements.")

        result = self._evaluator(x.ravel())

        if ndim == 1:
            return result
        else:
            return result.reshape([-1, 1])


def make_scalar_hypersingular(operator_descriptor, fmm_interface, domain, dual_to_range):
    """Create an evaluator for scalar hypersingular operators."""
    import bempp_cl.api
    from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    npoints = get_number_of_quad_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular)

    source_map = domain.map_to_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular)
    target_map = dual_to_range.map_to_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular, return_transpose=True)

    singular_part = operator_descriptor.singular_part.weak_form().to_sparse()

    source_normals = get_normals(domain, npoints)
    target_normals = get_normals(dual_to_range, npoints)

    source_curls, source_curls_trans = compute_p1_curl_transformation(
        domain, bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
    )

    if dual_to_range == domain:
        target_curls, target_curls_trans = source_curls, source_curls_trans
    else:
        target_curls, target_curls_trans = compute_p1_curl_transformation(
            dual_to_range, bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
        )

    def evaluate_laplace_hypersingular(x):
        """Evaluate the Laplace hypersingular kernel."""
        fmm_res0 = target_curls_trans[0] @ fmm_interface.evaluate(source_curls[0] @ x)[:, 0]
        fmm_res1 = target_curls_trans[1] @ fmm_interface.evaluate(source_curls[1] @ x)[:, 0]
        fmm_res2 = target_curls_trans[2] @ fmm_interface.evaluate(source_curls[2] @ x)[:, 0]

        return fmm_res0 + fmm_res1 + fmm_res2 + singular_part @ x

    def evaluate_helmholtz_hypersingular(x):
        """Evaluate the Helmholtz hypersingular kernel."""
        wavenumber = operator_descriptor.options[0] + 1j * operator_descriptor.options[1]
        x_transformed = source_map @ x

        fmm_res0 = target_curls_trans[0] @ fmm_interface.evaluate(source_curls[0] @ x)[:, 0]
        fmm_res1 = target_curls_trans[1] @ fmm_interface.evaluate(source_curls[1] @ x)[:, 0]
        fmm_res2 = target_curls_trans[2] @ fmm_interface.evaluate(source_curls[2] @ x)[:, 0]

        first_part = fmm_res0 + fmm_res1 + fmm_res2

        fmm_n1 = target_normals[:, 0] * fmm_interface.evaluate(source_normals[:, 0] * x_transformed)[:, 0]
        fmm_n2 = target_normals[:, 1] * fmm_interface.evaluate(source_normals[:, 1] * x_transformed)[:, 0]
        fmm_n3 = target_normals[:, 2] * fmm_interface.evaluate(source_normals[:, 2] * x_transformed)[:, 0]

        second_part = target_map @ (fmm_n1 + fmm_n2 + fmm_n3)

        return first_part - wavenumber * wavenumber * second_part + singular_part @ x

    def evaluate_modified_helmholtz_hypersingular(x):
        """Evaluate the modified Helmholtz hypersingular kernel."""
        wavenumber = operator_descriptor.options[0]
        x_transformed = source_map @ x

        fmm_res0 = target_curls_trans[0] @ fmm_interface.evaluate(source_curls[0] @ x)[:, 0]
        fmm_res1 = target_curls_trans[1] @ fmm_interface.evaluate(source_curls[1] @ x)[:, 0]
        fmm_res2 = target_curls_trans[2] @ fmm_interface.evaluate(source_curls[2] @ x)[:, 0]

        first_part = fmm_res0 + fmm_res1 + fmm_res2

        fmm_n1 = target_normals[:, 0] * fmm_interface.evaluate(source_normals[:, 0] * x_transformed)[:, 0]
        fmm_n2 = target_normals[:, 1] * fmm_interface.evaluate(source_normals[:, 1] * x_transformed)[:, 0]
        fmm_n3 = target_normals[:, 2] * fmm_interface.evaluate(source_normals[:, 2] * x_transformed)[:, 0]

        second_part = target_map @ (fmm_n1 + fmm_n2 + fmm_n3)

        return first_part + wavenumber * wavenumber * second_part + singular_part @ x

    if operator_descriptor.identifier == "laplace_hypersingular_boundary":
        return evaluate_laplace_hypersingular

    if operator_descriptor.identifier == "helmholtz_hypersingular_boundary":
        return evaluate_helmholtz_hypersingular

    if operator_descriptor.identifier == "modified_helmholtz_hypersingular_boundary":
        return evaluate_modified_helmholtz_hypersingular


def make_default_scalar(operator_descriptor, fmm_interface, domain, dual_to_range):
    """Create an evaluator for scalar operators."""
    import bempp_cl.api
    from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    npoints = get_number_of_quad_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular)

    source_map = domain.map_to_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular)
    target_map = dual_to_range.map_to_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular, return_transpose=True)

    singular_part = operator_descriptor.singular_part.weak_form().to_sparse()

    source_normals = get_normals(domain, npoints)
    target_normals = get_normals(dual_to_range, npoints)

    def evaluate_single_layer(x):
        """Actually evaluate single layer."""
        x_transformed = source_map @ x
        fmm_res = fmm_interface.evaluate(x_transformed)[:, 0]
        return target_map @ fmm_res + singular_part @ x

    def evaluate_adjoint_double_layer(x):
        """Actually evaluate adjoint double layer."""
        import numpy as np

        x_transformed = source_map @ x
        fmm_res = np.sum(fmm_interface.evaluate(x_transformed)[:, 1:] * target_normals, axis=1)
        return target_map @ fmm_res + singular_part @ x

    def evaluate_double_layer(x):
        """Actually evaluate double layer."""
        x_transformed = source_map @ x

        fmm_res1 = fmm_interface.evaluate(source_normals[:, 0] * x_transformed)[:, 1]
        fmm_res2 = fmm_interface.evaluate(source_normals[:, 1] * x_transformed)[:, 2]
        fmm_res3 = fmm_interface.evaluate(source_normals[:, 2] * x_transformed)[:, 3]

        fmm_res = -(fmm_res1 + fmm_res2 + fmm_res3)

        return target_map @ fmm_res + singular_part @ x

    if "single" in operator_descriptor.identifier:
        return evaluate_single_layer
    elif "adjoint_double" in operator_descriptor.identifier:
        return evaluate_adjoint_double_layer
    elif "double" in operator_descriptor.identifier:
        return evaluate_double_layer
    else:
        raise ValueError("Could not recognise identifier string.")


def get_normals(space, npoints):
    """Get the normal vectors on the quadrature points."""
    import numpy as np

    grid = space.grid
    number_of_elements = grid.number_of_elements

    normals = np.empty((npoints * number_of_elements, 3), dtype="float64")
    for element in range(number_of_elements):
        for n in range(npoints):
            normals[npoints * element + n, :] = grid.normals[element] * space.normal_multipliers[element]
    return normals


def compute_p1_curl_transformation(space, quadrature_order):
    """
    Compute the transformation of P1 space coefficients to surface curl values.

    Returns two lists, curl_transforms and curl_transforms_transpose. The jth matrix
    in curl_transforms is the map from P1 function space coefficients (or extended space
    built upon P1 type spaces) to the jth component of the surface curl evaluated at the
    quadrature points, multiplied with the quadrature weights and integration element. The
    list curl_transforms_transpose contains the transpose of these matrices.
    """
    from bempp_cl.api.integration.triangle_gauss import rule
    from scipy.sparse import coo_matrix
    from scipy.sparse.linalg import aslinearoperator

    grid_data = space.grid.data("double")
    number_of_elements = space.grid.number_of_elements
    quad_points, weights = rule(quadrature_order)
    npoints = len(weights)
    dof_count = space.localised_space.grid_dof_count

    data, iind, jind = compute_p1_curl_transformation_impl(
        grid_data,
        space.support_elements,
        space.normal_multipliers,
        quad_points,
        weights,
    )

    curl_transforms = []
    curl_transforms_transpose = []

    for index in range(3):
        curl_transforms.append(
            aslinearoperator(
                coo_matrix(
                    (data[index, :], (iind, jind)),
                    shape=(npoints * number_of_elements, dof_count),
                ).tocsr()
            )
            @ aslinearoperator(space.map_to_localised_space)
            @ aslinearoperator(space.dof_transformation)
        )
        curl_transforms_transpose.append(
            aslinearoperator(space.dof_transformation.T)
            @ aslinearoperator(space.map_to_localised_space.T)
            @ aslinearoperator(
                coo_matrix(
                    (data[index, :], (jind, iind)),
                    shape=(dof_count, npoints * number_of_elements),
                ).tocsr()
            )
        )

    return curl_transforms, curl_transforms_transpose


@_numba.njit
def compute_p1_curl_transformation_impl(grid_data, support_elements, normal_multipliers, quad_points, weights):
    """Implement the curl transformation."""
    number_of_quad_points = quad_points.shape[1]
    number_of_support_elements = len(support_elements)

    nlocal = 3 * number_of_quad_points

    data = _np.empty((3, nlocal * number_of_support_elements), dtype=_np.float64)
    jind = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)
    iind = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)

    reference_values = _np.array([[-1, 1, 0], [-1, 0, 1]], dtype=_np.float64)

    for element_index, element in enumerate(support_elements):
        surface_gradient = grid_data.jac_inv_trans[element] @ reference_values
        for function_index in range(3):
            surface_curl = normal_multipliers[element] * _np.cross(
                grid_data.normals[element], surface_gradient[:, function_index]
            )

            for point_index in range(number_of_quad_points):
                index = nlocal * element_index + function_index * number_of_quad_points + point_index
                data[:, index] = surface_curl * (weights[point_index] * grid_data.integration_elements[element])
                iind[index] = number_of_quad_points * element_index + point_index
                jind[index] = 3 * element_index + function_index

    return (data, iind, jind)


def compute_rwg_basis_transform(space, quadrature_order):
    """Compute the transformation matrices for RWG basis functions."""
    from bempp_cl.api.integration.triangle_gauss import rule
    from scipy.sparse import coo_matrix
    from scipy.sparse.linalg import aslinearoperator
    from bempp_cl.api.space.shapesets import _rwg0_shapeset_evaluate
    from bempp_cl.api.space.maxwell_spaces import _numba_rwg0_evaluate

    grid_data = space.grid.data("double")
    number_of_elements = space.grid.number_of_elements
    quad_points, weights = rule(quadrature_order)
    npoints = len(weights)
    dof_count = space.localised_space.grid_dof_count

    shapeset_evaluate = _rwg0_shapeset_evaluate
    basis_eval = _numba_rwg0_evaluate

    data, iind, jind = compute_rwg_basis_transform_impl(
        grid_data,
        shapeset_evaluate,
        basis_eval,
        space.support_elements,
        space.localised_space.local_multipliers,
        space.normal_multipliers,
        quad_points,
        weights,
    )

    basis_transforms = []
    basis_transforms_transpose = []

    for index in range(3):
        basis_transforms.append(
            aslinearoperator(
                coo_matrix(
                    (data[index, :], (iind, jind)),
                    shape=(npoints * number_of_elements, dof_count),
                ).tocsr()
            )
            @ aslinearoperator(space.map_to_localised_space)
            @ aslinearoperator(space.dof_transformation)
        )
        basis_transforms_transpose.append(
            aslinearoperator(space.dof_transformation.T)
            @ aslinearoperator(space.map_to_localised_space.T)
            @ aslinearoperator(
                coo_matrix(
                    (data[index, :], (jind, iind)),
                    shape=(dof_count, npoints * number_of_elements),
                ).tocsr()
            )
        )

    return basis_transforms, basis_transforms_transpose


@_numba.njit
def compute_rwg_basis_transform_impl(
    grid_data,
    shapeset_evaluate,
    basis_evaluate,
    support_elements,
    local_multipliers,
    normal_multipliers,
    quad_points,
    weights,
):
    """Implement the RWG basis transformation."""
    number_of_quad_points = quad_points.shape[1]
    number_of_support_elements = len(support_elements)

    nlocal = 3 * number_of_quad_points

    data = _np.empty((3, nlocal * number_of_support_elements), dtype=_np.float64)
    jind = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)
    iind = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)
    basis_values = _np.empty((3, 3, number_of_quad_points), dtype=_np.float64)

    for element_index, element in enumerate(support_elements):
        basis_values = basis_evaluate(
            element,
            shapeset_evaluate,
            quad_points,
            grid_data,
            local_multipliers,
            normal_multipliers,
        )
        for function_index in range(3):
            for point_index in range(number_of_quad_points):
                index = nlocal * element_index + function_index * number_of_quad_points + point_index
                data[:, index] = basis_values[:, function_index, point_index] * (
                    weights[point_index] * grid_data.integration_elements[element]
                )
                iind[index] = number_of_quad_points * element_index + point_index
                jind[index] = 3 * element_index + function_index

    return (data, iind, jind)


def compute_rwg_div_transform(space, quadrature_order):
    """Compute the div transformation matrices for RWG basis functions."""
    from bempp_cl.api.integration.triangle_gauss import rule
    from bempp_cl.api.space.shapesets import _rwg0_shapeset_evaluate
    from bempp_cl.api.space.maxwell_spaces import _numba_rwg0_evaluate
    from scipy.sparse import coo_matrix
    from scipy.sparse.linalg import aslinearoperator

    grid_data = space.grid.data("double")
    number_of_elements = space.grid.number_of_elements
    quad_points, weights = rule(quadrature_order)
    npoints = len(weights)
    dof_count = space.localised_space.grid_dof_count

    shapeset_evaluate = _rwg0_shapeset_evaluate
    basis_eval = _numba_rwg0_evaluate

    data, iind, jind = compute_rwg_div_transform_impl(
        grid_data,
        shapeset_evaluate,
        basis_eval,
        space.support_elements,
        space.localised_space.local_multipliers,
        space.normal_multipliers,
        quad_points,
        weights,
    )

    return (
        aslinearoperator(
            coo_matrix(
                (data, (iind, jind)),
                shape=(npoints * number_of_elements, dof_count),
            ).tocsr()
        )
        @ aslinearoperator(space.map_to_localised_space)
        @ aslinearoperator(space.dof_transformation),
        aslinearoperator(space.dof_transformation.T)
        @ aslinearoperator(space.map_to_localised_space.T)
        @ aslinearoperator(
            coo_matrix(
                (data, (jind, iind)),
                shape=(dof_count, npoints * number_of_elements),
            ).tocsr()
        ),
    )


@_numba.njit
def compute_rwg_div_transform_impl(
    grid_data,
    shapeset_evaluate,
    basis_evaluate,
    support_elements,
    local_multipliers,
    normal_multipliers,
    quad_points,
    weights,
):
    """Implement the RWG basis div transformation."""
    number_of_quad_points = quad_points.shape[1]
    number_of_support_elements = len(support_elements)

    nlocal = 3 * number_of_quad_points

    data = _np.empty(nlocal * number_of_support_elements, dtype=_np.float64)
    jind = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)
    iind = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)

    for element_index, element in enumerate(support_elements):
        edge_lengths = _np.empty(3, dtype=_np.float64)
        edge_lengths[0] = _np.linalg.norm(
            grid_data.vertices[:, grid_data.elements[0, element]]
            - grid_data.vertices[:, grid_data.elements[1, element]]
        )
        edge_lengths[1] = _np.linalg.norm(
            grid_data.vertices[:, grid_data.elements[2, element]]
            - grid_data.vertices[:, grid_data.elements[0, element]]
        )
        edge_lengths[2] = _np.linalg.norm(
            grid_data.vertices[:, grid_data.elements[1, element]]
            - grid_data.vertices[:, grid_data.elements[2, element]]
        )

        for function_index in range(3):
            for point_index in range(number_of_quad_points):
                index = nlocal * element_index + function_index * number_of_quad_points + point_index
                data[index] = 2.0 * edge_lengths[function_index] * (weights[point_index])
                iind[index] = number_of_quad_points * element_index + point_index
                jind[index] = 3 * element_index + function_index

    return (data, iind, jind)


def make_default_scalar_potential(operator_descriptor, fmm_interface, space):
    """Make a scalar potential operator."""
    import bempp_cl.api
    from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    npoints = get_number_of_quad_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular)
    source_map = space.map_to_points(bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular)
    source_normals = get_normals(space, npoints)

    def evaluate_single_layer(x):
        """Evaluate the single-layer operator."""
        return fmm_interface.evaluate(source_map @ x)[:, 0].reshape([1, -1])

    def evaluate_double_layer(x):
        """Evaluate the double-layer operator."""
        x_transformed = source_map @ x

        fmm0 = fmm_interface.evaluate(source_normals[:, 0] * x_transformed)[:, 1]
        fmm1 = fmm_interface.evaluate(source_normals[:, 1] * x_transformed)[:, 2]
        fmm2 = fmm_interface.evaluate(source_normals[:, 2] * x_transformed)[:, 3]

        return -(fmm0 + fmm1 + fmm2).reshape([1, -1])

    if "single" in operator_descriptor.identifier:
        return evaluate_single_layer
    elif "double" in operator_descriptor.identifier:
        return evaluate_double_layer


def make_maxwell_electric_field_boundary(operator_descriptor, fmm_interface, domain, dual_to_range):
    """Make a Maxwell electric field boundary operator."""
    import bempp_cl.api

    # from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    wavenumber = operator_descriptor.options[0] + 1j * operator_descriptor.options[1]
    order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
    domain_rwg_map, dual_rwg_map = compute_rwg_basis_transform(domain, order)
    domain_div_map, dual_div_map = compute_rwg_div_transform(domain, order)
    if domain != dual_to_range:
        _, dual_rwg_map = compute_rwg_basis_transform(dual_to_range, order)
        _, dual_div_map = compute_rwg_div_transform(dual_to_range, order)
    singular_part = operator_descriptor.singular_part.weak_form().to_sparse()

    def evaluate(x):
        """Evaluate the electric field operator."""
        result = _np.zeros(dual_to_range.global_dof_count, dtype=_np.complex128)

        for index in range(3):
            result += dual_rwg_map[index] @ fmm_interface.evaluate(domain_rwg_map[index] @ x)[:, 0]

        result *= -1j * wavenumber
        result -= 1 / (1j * wavenumber) * (dual_div_map @ fmm_interface.evaluate(domain_div_map @ x))[:, 0]
        return result + singular_part @ x

    return evaluate


def make_maxwell_magnetic_field_boundary(operator_descriptor, fmm_interface, domain, dual_to_range):
    """Make a Maxwell magnetic field boundary operator."""
    import bempp_cl.api

    # from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    # wavenumber = operator_descriptor.options[0]
    order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
    domain_rwg_map, dual_rwg_map = compute_rwg_basis_transform(domain, order)
    domain_div_map, dual_div_map = compute_rwg_div_transform(domain, order)
    if domain != dual_to_range:
        _, dual_rwg_map = compute_rwg_basis_transform(dual_to_range, order)
        _, dual_div_map = compute_rwg_div_transform(dual_to_range, order)

    singular_part = operator_descriptor.singular_part.weak_form().to_sparse()

    def evaluate(x):
        """Evaluate the magnetic field operator."""
        result = _np.zeros(dual_to_range.global_dof_count, dtype=_np.complex128)

        vals = [
            fmm_interface.evaluate(domain_rwg_map[0] @ x)[:, 1:],
            fmm_interface.evaluate(domain_rwg_map[1] @ x)[:, 1:],
            fmm_interface.evaluate(domain_rwg_map[2] @ x)[:, 1:],
        ]

        # Now compute the curl
        curl_val = _np.hstack(
            [
                (vals[2][:, 1] - vals[1][:, 2]).reshape(-1, 1),
                (vals[0][:, 2] - vals[2][:, 0]).reshape(-1, 1),
                (vals[1][:, 0] - vals[0][:, 1]).reshape(-1, 1),
            ]
        )

        # Finally, compute the negative inner product
        result = -(
            dual_rwg_map[0] @ curl_val[:, 0] + dual_rwg_map[1] @ curl_val[:, 1] + dual_rwg_map[2] @ curl_val[:, 2]
        )

        return result + singular_part @ x

    return evaluate


def make_maxwell_electric_field_potential(operator_descriptor, fmm_interface, space):
    """Make a Maxwell electric field potential operator."""
    import bempp_cl.api

    # from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    wavenumber = operator_descriptor.options[0] + 1j * operator_descriptor.options[1]
    order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
    rwg_map, rwg_map_trans = compute_rwg_basis_transform(space, order)
    div_map, div_map_trans = compute_rwg_div_transform(space, order)

    def evaluate(x):
        """Evaluate the potential operator."""
        res = (
            1j
            * wavenumber
            * _np.vstack(
                [
                    fmm_interface.evaluate(rwg_map[0] @ x)[:, 0],
                    fmm_interface.evaluate(rwg_map[1] @ x)[:, 0],
                    fmm_interface.evaluate(rwg_map[2] @ x)[:, 0],
                ]
            )
            - 1.0 / (1j * wavenumber) * fmm_interface.evaluate(div_map @ x)[:, 1:].T
        )

        return res

    return evaluate


def make_maxwell_magnetic_field_potential(operator_descriptor, fmm_interface, space):
    """Make a Maxwell magnetic field potential operator."""
    import bempp_cl.api

    # from bempp_cl.api.integration.triangle_gauss import get_number_of_quad_points

    # wavenumber = operator_descriptor.options[0]
    order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular
    rwg_map, rwg_map_trans = compute_rwg_basis_transform(space, order)
    div_map, div_map_trans = compute_rwg_div_transform(space, order)

    def evaluate(x):
        """Evaluate the potential operator."""
        vals = [
            fmm_interface.evaluate(rwg_map[0] @ x)[:, 1:],
            fmm_interface.evaluate(rwg_map[1] @ x)[:, 1:],
            fmm_interface.evaluate(rwg_map[2] @ x)[:, 1:],
        ]

        # Now compute the curl
        curl_val = _np.vstack(
            [
                (vals[2][:, 1] - vals[1][:, 2]),
                (vals[0][:, 2] - vals[2][:, 0]),
                (vals[1][:, 0] - vals[0][:, 1]),
            ]
        )
        return curl_val

    return evaluate


def clear_fmm_cache():
    """Clean the FMM cache."""
    global _FMM_CACHE
    global _FMM_POTENTIAL_CACHE

    _FMM_CACHE = {}
    _FMM_POTENTIAL_CACHE = {}
